# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rdflib import graph
from rdflib.namespace import SKOS

from stwfsapy import thesaurus as t


def test_extract_broader(mocker):
    g = graph.Graph()
    spy = mocker.spy(g, "subject_objects")
    t.extract_relation_by_uri(g, SKOS.broader, False)
    spy.assert_called_once_with(SKOS.broader)


def test_can_reverse(mocker):
    def mock_iter(a):
        yield (1, 3)
        yield (2, 4)
    g = graph.Graph()
    mocker.patch.object(g, 'subject_objects', mock_iter)
    res = t.extract_relation_by_uri(g, SKOS.broader, True)
    assert list(res) == [(3, 1), (4, 2)]



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from stwfsapy import thesaurus as t
import stwfsapy.tests.thesaurus.common as c


@pytest.fixture
def multi_lang_tuples():
    return [
        (c.concept_ref_printed, c.concept_prefLabel_printed_en),
        (c.concept_ref_printed, c.concept_prefLabel_printed_de),
        (c.concept_ref_printed, c.concept_prefLabel_printed_missing)
    ]


def test_filter_multiple_languages(multi_lang_tuples):
    langs = {"de", "en"}
    filtered = list(t._filter_by_langs(multi_lang_tuples, langs))
    assert len(filtered) == 2
    for ref, label in filtered:
        assert label.language in langs


def test_filter_single_language(multi_lang_tuples):
    lang = "de"
    langs = {lang}
    filtered = list(t._filter_by_langs(multi_lang_tuples, langs))
    assert len(filtered) == 1
    for ref, label in filtered:
        assert label.language == lang


def test_filter_none_language(multi_lang_tuples):
    lang = None
    langs = {lang}
    filtered = list(t._filter_by_langs(multi_lang_tuples, langs))
    assert len(filtered) == 1
    for ref, label in filtered:
        assert label.language == lang



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from stwfsapy.automata import nfa
from stwfsapy.tests.automata.data import symbol0


@pytest.fixture
def two_state_graph():
    graph = nfa.Nfa()
    graph.add_state()
    graph.add_state()
    return graph


@pytest.fixture
def epsilon_tree():
    graph = nfa.Nfa()
    for _ in range(17):
        graph.add_state()
    for i in range(8):
        graph.add_empty_transition(i, 2*i+1)
        graph.add_empty_transition(i, 2*i+2)
        graph.add_symbol_transition(i, 2*i+1, symbol0)
        graph.add_symbol_transition(i, 2*i+2, symbol0)
        graph.add_non_word_char_transition(i, 2*i+1)
        graph.add_non_word_char_transition(i, 2*i+2)
    return graph


@pytest.fixture
def epsilon_circle(epsilon_tree):
    epsilon_tree.add_empty_transition(16, 3)
    return epsilon_tree


def test_add_state(two_state_graph):
    two_state_graph.add_state()
    assert len(two_state_graph.states) == 3


def test_add_start(two_state_graph):
    two_state_graph.add_start(1)
    assert 1 in two_state_graph.starts


def test_add_acceptance(two_state_graph):
    acc = "Acceptance!"
    two_state_graph.add_acceptance(1, acc)
    assert acc in two_state_graph.states[1].accepts


def test_add_symbol_transition(two_state_graph):
    two_state_graph.add_symbol_transition(0, 1, symbol0)
    assert 1 in two_state_graph.states[0].symbol_transitions[symbol0]


def test_add_empty_transition(two_state_graph):
    two_state_graph.add_empty_transition(0, 1)
    assert 1 in two_state_graph.states[0].empty_transitions
    assert 0 in two_state_graph.states[1].incoming_empty_transitions


def test_add_non_word_char_transition(two_state_graph):
    two_state_graph.add_non_word_char_transition(0, 1)
    assert 1 in two_state_graph.states[0].non_word_char_transitions


def test_remove_epsilons(epsilon_tree):
    epsilon_tree.remove_empty_transitions()
    state = epsilon_tree.states[0]
    assert len(state.empty_transitions) == 0
    assert len(state.incoming_empty_transitions) == 0
    for idx in range(1, 16):
        assert idx in epsilon_tree.states[0].symbol_transitions[symbol0]
        assert idx in epsilon_tree.states[0].non_word_char_transitions
        state = epsilon_tree.states[idx]
        assert len(state.empty_transitions) == 0
        assert len(state.incoming_empty_transitions) == 0


def test_recognizes_empty_loops(epsilon_circle):
    with pytest.raises(Exception) as exc:
        epsilon_circle.remove_empty_transitions()
    assert exc.value.args[0] == "There is an empty transition loop in the NFA."



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from stwfsapy import thesaurus as t
from stwfsapy.tests.thesaurus import common as c


def test_filter_subject_tuples_from_set(tuples):
    res = list(
        t.filter_subject_tuples_from_set(tuples, {c.concept_ref_printed}))
    assert len(res) == 2
    assert res[0][0] == c.concept_ref_printed
    assert res[1][0] == c.concept_ref_printed


def test_filter_refs_from_set_complement():
    res = list(
        t._filter_refs_from_set_complement(
            [c.concept_ref_printed, c.concept_ref_media],
            {c.concept_ref_media}
        )
    )
    assert res == [c.concept_ref_printed]



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from stwfsapy.automata import conversion as c
from stwfsapy.tests.automata.data import accept, symbol0, symbol1


def test_conversion(input_graph, mocker):
    converter = c.NfaToDfaConverter(input_graph)
    # Mainly interested in the end result.
    # Still mock to see that there are no unknown elements
    add_spy = mocker.spy(converter.dfa, "add_state")
    symbol_transition_spy = mocker.spy(converter.dfa, "set_symbol_transition")
    nwc_transition_spy = mocker.spy(
        converter.dfa,
        "set_non_word_char_transition")
    acceptance_spy = mocker.spy(converter.dfa, "add_acceptances")
    result = converter.start_conversion()
    assert converter.queue.qsize() == 0
    assert len(converter.state_cache) == 6
    assert len(converter.state_represents) == 6
    # One less call than states in the graph,
    # because the initial state was added during construction.
    assert add_spy.call_count == 5
    assert nwc_transition_spy.call_count == 1
    assert symbol_transition_spy.call_count == 7
    assert acceptance_spy.call_count == 1
    state01 = converter.state_cache[frozenset([0, 1])]
    state02 = converter.state_cache[frozenset([0, 2])]
    assert result.states[state01].symbol_transitions[symbol0] == state02
    state3 = converter.state_cache[frozenset([3])]
    assert result.states[state01].symbol_transitions[symbol1] == state3
    assert result.states[state02].symbol_transitions[symbol1] == state3
    state24 = converter.state_cache[frozenset([2, 4])]
    assert result.states[state02].symbol_transitions[symbol0] == state24
    state5 = converter.state_cache[frozenset([5])]
    assert result.states[state24].symbol_transitions[symbol1] == state5
    state4 = converter.state_cache[frozenset([4])]
    assert result.states[state24].symbol_transitions[symbol0] == state4
    assert result.states[state4].symbol_transitions[symbol1] == state5
    assert result.states[state5].accepts == [accept]


def test_creates_new_state(input_graph):
    converter = c.NfaToDfaConverter(input_graph)
    new_set = frozenset([0, 2])
    converter._get_or_create_dfa_state(new_set)
    assert len(converter.dfa.states) == 2
    assert converter.queue.qsize() == 2
    assert len(converter.state_cache) == 2
    assert len(converter.state_represents) == 2
    assert converter.state_represents[1] == list(new_set)
    assert converter.state_cache[new_set] == 1


def test_retrieves_existing_state(input_graph):
    converter = c.NfaToDfaConverter(input_graph)
    new_set = frozenset(input_graph.starts)
    converter._get_or_create_dfa_state(new_set)
    assert len(converter.dfa.states) == 1
    assert converter.queue.qsize() == 1
    assert len(converter.state_cache) == 1
    assert len(converter.state_represents) == 1


def test_initialization(input_graph):
    converter = c.NfaToDfaConverter(input_graph)
    start_set = frozenset(input_graph.starts)
    assert len(converter.dfa.states) == 1
    assert converter.queue.qsize() == 1
    assert len(converter.state_cache) == 1
    assert len(converter.state_represents) == 1
    assert converter.state_represents[0] == list(start_set)
    assert converter.state_cache[start_set] == 0


def test_transition_collection(input_graph):
    converter = c.NfaToDfaConverter(input_graph)
    state_set = {0, 1, 3, 5}
    (
        symbol_transitions,
        non_word_char_transitions,
        accepts
    ) = converter._collect_nfa_transitions(state_set)
    for state_id in state_set:
        state = input_graph.states[state_id]
        for k, idxs in state.symbol_transitions.items():
            for idx in idxs:
                assert idx in symbol_transitions[k]
        for v in state.non_word_char_transitions:
            assert v in non_word_char_transitions
        for v in state.accepts:
            assert v in accepts
    for k, idxs in symbol_transitions.items():
        for idx in idxs:
            found = False
            for state in state_set:
                nfa_state = input_graph.states[state]
                if idx in nfa_state.symbol_transitions.get(k, []):
                    found = True
            assert found
    for v in non_word_char_transitions:
        found = False
        for state in state_set:
            if v in input_graph.states[state].non_word_char_transitions:
                found = True
        assert found
    for v in accepts:
        found = False
        for state in state_set:
            if v in input_graph.states[state].accepts:
                found = True
        assert found


def test_transition_creation(input_graph):
    converter = c.NfaToDfaConverter(input_graph)
    set13 = frozenset([1, 3])
    set24 = frozenset([2, 4])
    set5 = frozenset([5])
    acceptance = {accept}
    non_word_char_transitions = set(set24)
    symbol_transitions = {
        symbol0: set(set13),
        symbol1: set(set5)
    }
    converter._create_dfa_transitions(
        0,
        symbol_transitions,
        non_word_char_transitions,
        acceptance
    )
    state = converter.dfa.states[0]
    assert state.accepts == [accept]
    assert state.non_word_char_transition == converter.state_cache[set24]
    assert state.symbol_transitions[symbol0] == converter.state_cache[set13]
    assert state.symbol_transitions[symbol1] == converter.state_cache[set5]



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import stwfsapy.expansion as e
from inspect import signature

_name_abbreviation_fun = e._expand_abbreviation_with_punctuation_fun.__name__
_name_ampersand_fun = e._expand_ampersand_with_spaces_fun.__name__
_name_replacer = "replacer"

lower_braces_content = "foo"
lower_braces_string = "\\({}\\) bar".format(lower_braces_content)

_conf_keys = list(signature(e.collect_expansion_functions).parameters)
"""All parameters of the collect_expansion_functions method"""


def _create_config_map(lst):
    return {
        k: bool(int(v))
        for k, v
        in zip(_conf_keys, lst)}


def test_combination_0000():
    conf = "00000"
    conf_map = _create_config_map(conf)
    assert e.collect_expansion_functions(**conf_map) == [e.base_expansion]


def test_combination_0001():
    conf = "00010"
    conf_map = _create_config_map(conf)
    funs = e.collect_expansion_functions(**conf_map)
    assert len(funs) == 2
    assert funs[0] == e.base_expansion
    assert funs[1].__name__ == _name_abbreviation_fun


def test_combination_0010():
    conf = "00100"
    conf_map = _create_config_map(conf)
    funs = e.collect_expansion_functions(**conf_map)
    assert len(funs) == 2
    assert funs[0] == e.base_expansion
    assert funs[1].__name__ == _name_ampersand_fun


def test_combination_0100():
    conf = "01000"
    conf_map = _create_config_map(conf)
    funs = e.collect_expansion_functions(**conf_map)
    assert len(funs) == 2
    assert funs[0] == e.base_expansion
    assert funs[1].__name__ == _name_replacer
    assert funs[1](lower_braces_string) == lower_braces_content


def test_combination_1000():
    conf = "10000"
    conf_map = _create_config_map(conf)
    funs = e.collect_expansion_functions(**conf_map)
    assert len(funs) == 2
    assert funs[0] == e.base_expansion
    assert funs[1].__name__ == _name_replacer
    assert funs[1](lower_braces_string) == lower_braces_string


def test_combination_0011():
    conf = "00110"
    conf_map = _create_config_map(conf)
    funs = e.collect_expansion_functions(**conf_map)
    assert len(funs) == 3
    assert funs[0] == e.base_expansion
    assert funs[1].__name__ == _name_ampersand_fun
    assert funs[2].__name__ == _name_abbreviation_fun


def test_combination_0101():
    conf = "01010"
    conf_map = _create_config_map(conf)
    funs = e.collect_expansion_functions(**conf_map)
    assert len(funs) == 3
    assert funs[0] == e.base_expansion
    assert funs[1].__name__ == _name_replacer
    assert funs[2].__name__ == _name_abbreviation_fun
    assert funs[1](lower_braces_string) == lower_braces_content


def test_combination_1001():
    conf = "10010"
    conf_map = _create_config_map(conf)
    funs = e.collect_expansion_functions(**conf_map)
    assert len(funs) == 3
    assert funs[0] == e.base_expansion
    assert funs[1].__name__ == _name_replacer
    assert funs[2].__name__ == _name_abbreviation_fun
    assert funs[1](lower_braces_string) == lower_braces_string


def test_combination_0110():
    conf = "01100"
    conf_map = _create_config_map(conf)
    funs = e.collect_expansion_functions(**conf_map)
    assert len(funs) == 3
    assert funs[0] == e.base_expansion
    assert funs[1].__name__ == _name_replacer
    assert funs[2].__name__ == _name_ampersand_fun
    assert funs[1](lower_braces_string) == lower_braces_content


def test_combination_1010():
    conf = "10100"
    conf_map = _create_config_map(conf)
    funs = e.collect_expansion_functions(**conf_map)
    assert len(funs) == 3
    assert funs[0] == e.base_expansion
    assert funs[1].__name__ == _name_replacer
    assert funs[2].__name__ == _name_ampersand_fun
    assert funs[1](lower_braces_string) == lower_braces_string


def test_combination_1100():
    conf = "11000"
    conf_map = _create_config_map(conf)
    funs = e.collect_expansion_functions(**conf_map)
    assert len(funs) == 3
    assert funs[0] == e.base_expansion
    assert funs[1].__name__ == _name_replacer
    assert funs[2].__name__ == _name_replacer
    assert funs[1](lower_braces_string) == lower_braces_string
    assert funs[2](lower_braces_string) == lower_braces_content


def test_combination_0111():
    conf = "01110"
    conf_map = _create_config_map(conf)
    funs = e.collect_expansion_functions(**conf_map)
    assert len(funs) == 4
    assert funs[0] == e.base_expansion
    assert funs[1].__name__ == _name_replacer
    assert funs[2].__name__ == _name_ampersand_fun
    assert funs[3].__name__ == _name_abbreviation_fun
    assert funs[1](lower_braces_string) == lower_braces_content


def test_combination_1011():
    conf = "10110"
    conf_map = _create_config_map(conf)
    funs = e.collect_expansion_functions(**conf_map)
    assert len(funs) == 4
    assert funs[0] == e.base_expansion
    assert funs[1].__name__ == _name_replacer
    assert funs[2].__name__ == _name_ampersand_fun
    assert funs[3].__name__ == _name_abbreviation_fun
    assert funs[1](lower_braces_string) == lower_braces_string


def test_combination_1101():
    conf = "11010"
    conf_map = _create_config_map(conf)
    funs = e.collect_expansion_functions(**conf_map)
    assert len(funs) == 4
    assert funs[0] == e.base_expansion
    assert funs[1].__name__ == _name_replacer
    assert funs[2].__name__ == _name_replacer
    assert funs[3].__name__ == _name_abbreviation_fun
    assert funs[1](lower_braces_string) == lower_braces_string
    assert funs[2](lower_braces_string) == lower_braces_content


def test_combination_1110():
    conf = "11100"
    conf_map = _create_config_map(conf)
    funs = e.collect_expansion_functions(**conf_map)
    assert len(funs) == 4
    assert funs[0] == e.base_expansion
    assert funs[1].__name__ == _name_replacer
    assert funs[2].__name__ == _name_replacer
    assert funs[3].__name__ == _name_ampersand_fun
    assert funs[1](lower_braces_string) == lower_braces_string
    assert funs[2](lower_braces_string) == lower_braces_content


def test_combination_1111():
    conf = "11110"
    conf_map = _create_config_map(conf)
    funs = e.collect_expansion_functions(**conf_map)
    assert len(funs) == 5
    assert funs[0] == e.base_expansion
    assert funs[1].__name__ == _name_replacer
    assert funs[2].__name__ == _name_replacer
    assert funs[3].__name__ == _name_ampersand_fun
    assert funs[4].__name__ == _name_abbreviation_fun
    assert funs[1](lower_braces_string) == lower_braces_string
    assert funs[2](lower_braces_string) == lower_braces_content


def test_all_abbreviation_from_braces():
    conf = "11110"
    conf_map = _create_config_map(conf)
    funs = e.collect_expansion_functions(**conf_map)
    string = "GDP (gross domestic product)"
    for fun in funs:
        string = fun(string)
    assert string == "G\\.?D\\.?P\\.?"


def test_all_ampersand_from_braces():
    conf = "11110"
    conf_map = _create_config_map(conf)
    funs = e.collect_expansion_functions(**conf_map)
    string = "(R&D) research and discovery"
    for fun in funs:
        string = fun(string)
    assert string == "R ?& ?D"


def test_all_lower_braces():
    conf = "11110"
    conf_map = _create_config_map(conf)
    funs = e.collect_expansion_functions(**conf_map)
    string = "(lower) lower case explanation"
    for fun in funs:
        string = fun(string)
    assert string == "lower"



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from stwfsapy.frequency_features import FrequencyFeatures
from sklearn.exceptions import NotFittedError
import numpy as np
from math import log

import pytest

frequency_input = [
    ('cncpt_1',  [3, 4, 0, 2], 0),
    ('cncpt_2', [1, 8, 23, 12], 1),
    ('cncpt_1',  [12, 36, 25, 7], 0),
    ('cncpt_2', [8, 102, 17, 9, 20], 0),
    ('cncpt_3', [13], 1)
]


def test_handle_empty():
    features = FrequencyFeatures()
    features.fit([])
    assert features.transform([]).tolist() == np.zeros((0, 3)).tolist()


def test_fit():
    features = FrequencyFeatures()
    features.fit(frequency_input)
    keyset = {'cncpt_1', 'cncpt_2', 'cncpt_3'}
    assert set(features.idfs_.keys()) == keyset
    assert features.idfs_['cncpt_1'] == 0
    assert features.idfs_['cncpt_2'] == 0
    assert features.idfs_['cncpt_3'] == log(3/2)


def test_not_fitted():
    features = FrequencyFeatures()
    with pytest.raises(NotFittedError):
        features.transform([])


def test_transform():
    features = FrequencyFeatures()
    features.fit(frequency_input)
    data = [
        ('cncpt_2', [2, 5], 0),
        ('cncpt_3', [17, 11, 22], 1),
        ('cncpt_1', [13, 2, 1, 9, 8], 0),
        ('cncpt_3', [6], 0),
        ('unknown', [4, 3, 26, 39, 7], 1)
    ]
    transformed = features.transform(
        data
    )
    assert transformed.tolist() == [
        [2/5, 0, 0],
        [3/5, log(3/2), 3/5*log(3/2)],
        [5/11, 0, 0],
        [1/11, log(3/2), 1/11 * log(3/2)],
        [5/11, log(3), 5/11 * log(3)]
    ]



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rdflib import URIRef
from rdflib.term import Literal


test_type_thesaurus = URIRef("http://type.org/thesaurus")
test_type_concept = URIRef("http://type.org/concept")


test_thesaurus_ref_0 = URIRef("http://test.org/thesaurus/0")
test_thesaurus_ref_00 = URIRef("http://test.org/thesaurus/00")
test_thesaurus_ref_01 = URIRef("http://test.org/thesaurus/01")
test_thesaurus_ref_1 = URIRef("http://test.org/thesaurus/1")
test_thesaurus_ref_10 = URIRef("http://test.org/thesaurus/10")
test_thesaurus_ref_100 = URIRef("http://test.org/thesaurus/100")


test_thesauri = [
    test_thesaurus_ref_0,
    test_thesaurus_ref_00,
    test_thesaurus_ref_01,
    test_thesaurus_ref_1,
    test_thesaurus_ref_10,
    test_thesaurus_ref_100,

    ]
thsrs_broader = [
    (test_thesaurus_ref_00, test_thesaurus_ref_0),
    (test_thesaurus_ref_01, test_thesaurus_ref_0),
    (test_thesaurus_ref_10, test_thesaurus_ref_1),
    (test_thesaurus_ref_100, test_thesaurus_ref_10),
    ]

test_concept_uri_0_0 = "http://test.org/concept/0_0"
test_concept_uri_01_0 = "http://test.org/concept/01_0"
test_concept_uri_01_00 = "http://test.org/concept/01_00"
test_concept_uri_10_0 = "http://test.org/concept/10_0"
test_concept_uri_10_1 = "http://test.org/concept/10_1"
test_concept_uri_100_0 = "http://test.org/concept/100_0"
test_concept_uri_100_00 = "http://test.org/concept/100_00"
test_concept_uri_100_01 = "http://test.org/concept/100_01"
test_concept_uri_100_02 = "http://test.org/concept/100_02"

test_concept_ref_0_0 = URIRef(test_concept_uri_0_0)
test_concept_ref_01_0 = URIRef(test_concept_uri_01_0)
test_concept_ref_01_00 = URIRef(test_concept_uri_01_00)
test_concept_ref_10_0 = URIRef(test_concept_uri_10_0)
test_concept_ref_10_1 = URIRef(test_concept_uri_10_1)
test_concept_ref_100_0 = URIRef(test_concept_uri_100_0)
test_concept_ref_100_00 = URIRef(test_concept_uri_100_00)
test_concept_ref_100_01 = URIRef(test_concept_uri_100_01)
test_concept_ref_100_02 = URIRef(test_concept_uri_100_02)

test_labels = [
    (test_concept_ref_0_0, Literal("concept-0_0", lang="en")),
    (test_concept_ref_01_0, Literal("concept-01_0", lang="en")),
    (test_concept_ref_01_00, Literal("concept-01_00", lang="en")),
    (test_concept_ref_10_0, Literal("concept-10_0", lang="en")),
    (test_concept_ref_10_1, Literal("concept-10_1", lang="en")),
    (test_concept_ref_100_0, Literal("concept-100_0", lang="en")),
    (test_concept_ref_100_00, Literal("concept-100_00", lang="en")),
    (test_concept_ref_100_01, Literal("concept-100_01", lang="en")),
    (test_concept_ref_100_02, Literal("concept-100_02", lang="en")),
    ]

test_concepts = [
    test_concept_ref_0_0,
    test_concept_ref_01_0,
    test_concept_ref_01_00,
    test_concept_ref_10_0,
    test_concept_ref_10_1,
    test_concept_ref_100_0,
    test_concept_ref_100_00,
    test_concept_ref_100_01,
    test_concept_ref_100_02,
    ]

cncpt_broaders = [
    (test_concept_ref_0_0, test_thesaurus_ref_0),
    (test_concept_ref_01_0, test_thesaurus_ref_01),
    (test_concept_ref_01_00, test_thesaurus_ref_01),
    (test_concept_ref_01_00, test_concept_ref_01_0),
    (test_concept_ref_10_0, test_thesaurus_ref_10),
    (test_concept_ref_10_1, test_thesaurus_ref_10),
    (test_concept_ref_100_0, test_thesaurus_ref_100),
    (test_concept_ref_100_00, test_thesaurus_ref_100),
    (test_concept_ref_100_01, test_thesaurus_ref_100),
    (test_concept_ref_100_02, test_thesaurus_ref_100),
    (test_concept_ref_100_00, test_concept_ref_100_0),
    (test_concept_ref_100_01, test_concept_ref_100_0),
    (test_concept_ref_100_02, test_concept_ref_100_0),
    ]


doc_feature_data = [
    ('cncpt_1', 'f'*3, [3, 4, 0, 2], 0),
    ('cncpt_2', 'f'*6, [1, 8, 23, 12], 1),
    ('cncpt_1', 'f'*15, [12, 36, 25, 7], 0),
    ('cncpt_2', 'f'*12, [8, 102, 17, 9, 20], 0),
    ('cncpt_3', 'f'*70, [13], 1)
]



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from stwfsapy.expansion import base_expansion


def test_curly_open():
    input = 'some { thing'
    out = base_expansion(input)
    assert out == 'some \\{ thing'


def test_curly_close():
    input = 'some } thing'
    out = base_expansion(input)
    assert out == 'some \\} thing'


def test_curly_both():
    input = 'some {thing}'
    out = base_expansion(input)
    assert out == 'some \\{thing\\}'


def test_round_open():
    input = 'some ( thing'
    out = base_expansion(input)
    assert out == 'some \\( thing'


def test_round_close():
    input = 'some ) thing'
    out = base_expansion(input)
    assert out == 'some \\) thing'


def test_round_both():
    input = 'some (thing)'
    out = base_expansion(input)
    assert out == 'some \\(thing\\)'


def test_square_open():
    input = 'some [ thing'
    out = base_expansion(input)
    assert out == 'some \\[ thing'


def test_square_close():
    input = 'some ] thing'
    out = base_expansion(input)
    assert out == 'some \\] thing'


def test_square_both():
    input = 'some [thing]'
    out = base_expansion(input)
    assert out == 'some \\[thing\\]'


def test_asterisk():
    input = 'Something text with * inside'
    out = base_expansion(input)
    assert out == 'Something text with \\* inside'


def test_question_mark():
    input = 'Something text with ? inside'
    out = base_expansion(input)
    assert out == 'Something text with \\? inside'



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import stwfsapy.expansion as e

lower_string = "lower case"
with_hyphen_string = "WITH-HYPHEN"
upper_case_string = "UCS"


def test_does_not_change_lower():
    assert lower_string == \
        e._expand_abbreviation_with_punctuation_fun(lower_string)


def test_does_not_change_with_symbol():
    assert with_hyphen_string == \
        e._expand_abbreviation_with_punctuation_fun(with_hyphen_string)


def test_inserts_punctuation():
    result = e._expand_abbreviation_with_punctuation_fun(upper_case_string)
    assert len(result) == 4 * len(upper_case_string)
    for i in range(len(upper_case_string)):
        assert result[4*i] == upper_case_string[i]
        assert result[4*i+1] == '\\'
        assert result[4*i+2] == '.'
        assert result[4*i+3] == '?'



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class UnknownInputTypeException(Exception):
    def __init__(self) -> None:
        self.message = 'Unknown Input Format. Please specify' + (
            ' "content", "file" or "filename".')
        super().__init__(self.message)


def handle_content(txt):
    return txt


def handle_file(file):
    return file.read()


def handle_filename(f_pth):
    with open(f_pth) as f:
        return f.read()


def get_input_handler(input_type):
    if input_type == 'content':
        return handle_content
    elif input_type == 'file':
        return handle_file
    elif input_type == 'filename':
        return handle_filename
    raise UnknownInputTypeException()



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from stwfsapy import expansion as e
import common as c

replacement_fun_upper = e._replace_by_pattern_fun(
    e._upper_case_abbreviation_from_braces_expression)


def test_pattern_match():
    match = e._upper_case_abbreviation_from_braces_expression.match(
        c.test_string_upper
        )
    assert match.group(0) == c.test_string_upper
    assert match.group(1) == c.test_abbreviation_upper


def test_replacement():
    assert replacement_fun_upper(c.test_string_upper) == \
        c.test_abbreviation_upper


def test_no_replacement():
    assert replacement_fun_upper(c.test_string_any) == c.test_string_any


def test_no_match_when_not_at_beginning():
    test_string = " " + c.test_abbreviation_upper
    assert e._upper_case_abbreviation_from_braces_expression.search(
        test_string
    ) is None



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from stwfsapy.text_features import mk_text_features
from scipy.sparse.lil import lil_matrix
from stwfsapy import predictor as p
import stwfsapy.thesaurus as t
from stwfsapy.automata.dfa import Dfa
import stwfsapy.tests.common as c
from stwfsapy.automata.construction import ConstructionState
import pytest
from scipy.sparse import csr_matrix
import numpy as np
from sklearn.tree import DecisionTreeClassifier
from sklearn.compose import ColumnTransformer
from stwfsapy import case_handlers as handlers
from unittest.mock import call
from rdflib import Graph
from rdflib.namespace import SKOS
from rdflib.term import Literal

_doc_counts = [2, 4, 3]
_concepts = list(range(9, 18))
_concepts_with_text = [(c, str(c)) for c in _concepts]
_proto_preds = list(range(2, 11))
_predictions = np.array([[0, i/10] for i in _proto_preds])
_classifications = np.array([i % 2 for i in _proto_preds])
_concept_map = dict(zip(reversed(range(23)), range(23)))
_collection_result = [
    ([9, 10], [0.2, 0.3]),
    ([11, 12, 13, 14], [0.4, 0.5, 0.6, 0.7]),
    ([15, 16, 17], [0.8, 0.9, 1.0])
]
train_texts = [
    "concept-0_0",
    "nothing",
    "concept",
    "has Concept-100_00 in the middle",
    "concept-10_0 and concept-01_00 and concept-10_0",
    ]
train_labels = [
    [c.test_concept_ref_0_0],
    [c.test_concept_ref_01_0],
    [],
    [c.test_concept_ref_01_00],
    [c.test_concept_ref_10_0, c.test_concept_ref_01_00],
    ]
expected_matches = [
    ("9", [0], 1), ("9", [0], 0),
    ("11", [0, 1], 0), ("13", [0, 1, 2], 1),
    ("9", [0], 0), ("11", [0, 1], 1)]


def make_test_result_matrix(values):
    return csr_matrix((
        values,
        (
            [
                i
                for i, count
                in enumerate(_doc_counts)
                for _
                in range(count)],
            [22-c for c in _concepts])),
            shape=(3, 23))


@pytest.fixture
def patched_dfa(mocker):
    dfa = Dfa()

    def mock_search(text):
        for i in range(len(text)):
            for j in range(i+1):
                yield (str(i*2+9), text, j)

    mocker.patch.object(dfa, "search", mock_search)
    return dfa


@pytest.fixture
def mocked_predictor(mocker, mock_vectorizer):
    predictor = p.StwfsapyPredictor(None, None, None, None)
    predictor.concept_map_ = _concept_map
    predictor.match_and_extend = mocker.Mock(
        return_value=(_concepts_with_text, _doc_counts))
    predictor.text_vectorizer_ = mock_vectorizer
    predictor.text_features_ = mk_text_features()
    predictor.pipeline_ = mocker.MagicMock()
    predictor.pipeline_.predict_proba = mocker.Mock(return_value=_predictions)
    predictor.pipeline_.predict = mocker.Mock(return_value=_classifications)
    return predictor


@pytest.fixture
def no_match_predictor(mocker):
    predictor = p.StwfsapyPredictor(None, None, None, None)
    predictor.concept_map_ = _concept_map
    predictor.match_and_extend = mocker.Mock(
        return_value=([], [0, 0, 0]))
    return predictor


def mock_vec_transform(X):
    ret = lil_matrix((len(X), 5000))
    for idx, x in enumerate(X):
        ret[idx] = len(x)
    return ret


@pytest.fixture
def mock_vectorizer(mocker):
    mock_vec = mocker.MagicMock()
    mock_vec.transform = mock_vec_transform
    return mock_vec


@pytest.fixture
def case_graph():
    g = Graph()
    g.add((
        c.test_concept_ref_0_0,
        SKOS.prefLabel,
        Literal("Three word label")))
    return g


def test_result_collection():
    res = p.StwfsapyPredictor._collect_prediction_results(
        _predictions[:, 1],
        [c[0] for c in _concepts_with_text],
        _doc_counts
    )
    assert [(r[0], list(r[1])) for r in res] == _collection_result


def test_sparse_matrix_creation():
    predictor = p.StwfsapyPredictor(None, None, None, None)
    predictor.concept_map_ = _concept_map
    res = predictor._create_sparse_matrix(
        _predictions[:, 1],
        [c[0] for c in _concepts_with_text],
        _doc_counts
    )
    assert res.shape[0] == len(_doc_counts)
    assert res.shape[1] == 23
    for i, count in enumerate(_doc_counts):
        row = res.getrow(i)
        slice_start = sum(_doc_counts[:i])
        assert row.getnnz() == count
        # reverse slices because of mapping.
        assert list(row.nonzero()[1]) == list(reversed([
            22-i for i in _concepts[slice_start: slice_start+count]]))
        assert list(row.data) == list(reversed(
            _predictions[slice_start:slice_start+count, 1]))


def test_match_and_extend_with_truth_with_vec(patched_dfa, mock_vectorizer):
    predictor = p.StwfsapyPredictor(None, None, None, None, use_txt_vec=True)
    predictor.dfa_ = patched_dfa
    predictor.text_features_ = mk_text_features().fit([])
    predictor.text_vectorizer_ = mock_vectorizer
    in_txts = ["a", "bbb", "xx"]
    matches, ys = predictor.match_and_extend(
        in_txts,
        [[], [11, 14], [9]]
    )
    txt_indices = [
        idx
        for idx, txt
        in enumerate(in_txts)
        for _ in range(len(txt))
    ]
    for txt_idx, match, expected in zip(
            txt_indices,
            matches,
            expected_matches):
        check_fit_arg(
            predictor.text_vectorizer_,
            predictor.text_features_,
            in_txts[txt_idx],
            match,
            expected
        )
    assert ys == [0, 0, 1, 0, 1, 0]


def test_match_and_extend_without_truth_with_vec(patched_dfa, mock_vectorizer):
    predictor = p.StwfsapyPredictor(None, None, None, None, use_txt_vec=True)
    predictor.dfa_ = patched_dfa
    predictor.text_features_ = mk_text_features().fit([])
    predictor.text_vectorizer_ = mock_vectorizer
    in_txts = ["a", "bbb", "xx"]
    matches, counts = predictor.match_and_extend(in_txts)
    txt_indices = [
        idx
        for idx, txt
        in enumerate(in_txts)
        for _ in range(len(txt))
    ]
    for txt_idx, match, expected in zip(
            txt_indices,
            matches,
            expected_matches):
        check_fit_arg(
            predictor.text_vectorizer_,
            predictor.text_features_,
            in_txts[txt_idx],
            match,
            expected
        )
    assert counts == [1, 3, 2]


def test_init_and_fit_with_vec(full_graph, mocker):
    predictor = p.StwfsapyPredictor(
        full_graph,
        c.test_type_concept,
        c.test_type_thesaurus,
        SKOS.broader,
        use_txt_vec=True)
    spy_deprecated = mocker.spy(t, "extract_deprecated")
    spy_case = mocker.spy(handlers, 'title_case_handler')
    predictor._init()
    assert predictor.text_vectorizer_ is not None
    spy_deprecated.assert_called_once_with(full_graph)
    assert len(spy_case.mock_calls) == len(c.test_labels)
    for _, label in c.test_labels:
        assert call(label.toPython()) in spy_case.mock_calls
    assert isinstance(
        predictor.pipeline_.named_steps['Classifier'],
        DecisionTreeClassifier)
    combined = predictor.pipeline_.named_steps['Combined Features']
    assert isinstance(
        combined,
        ColumnTransformer)
    assert len(combined.transformers) == 5
    assert combined.transformers[0][0] == 'Thesaurus Features'
    assert combined.transformers[1][0] == 'Text Features'
    assert combined.transformers[2][0] == 'Position Features'
    assert combined.transformers[3][0] == 'Frequency Features'
    assert combined.transformers[4][0] == 'Text Vector'
    assert combined.transformers[0][1].thesaurus_relation == SKOS.broader
    spy_fit_pipeline = mocker.spy(predictor.pipeline_, "fit")
    spy_fit_vec = mocker.spy(predictor.text_vectorizer_, "fit")
    predictor._fit_after_init(train_texts, y=train_labels)
    spy_fit_pipeline.assert_called_once()

    pipe_fit_arg_list = spy_fit_pipeline.call_args_list[0]
    assert len(pipe_fit_arg_list) == 2
    pipe_fit_args = pipe_fit_arg_list[0]
    assert pipe_fit_arg_list[1].get('y') == [1, 0, 1, 1]
    expected_features = [
            (c.test_concept_uri_0_0, [0], 1),
            (
                c.test_concept_uri_100_00,
                [4],
                1),
            (c.test_concept_uri_10_0, [0, 35], 0),
            (c.test_concept_uri_01_00, [17], 1),
            ]
    assert len(expected_features) == len(pipe_fit_args[0])
    txt_indices = [0, 3, 4, 4]
    for txt_idx, actual, expected in zip(
            txt_indices,
            pipe_fit_args[0],
            expected_features):
        check_fit_arg(
            predictor.text_vectorizer_,
            predictor.text_features_,
            train_texts[txt_idx],
            actual,
            expected)

    spy_fit_vec.assert_called_once_with(
        train_texts
    )


def test_match_and_extend_with_truth_without_vec(patched_dfa):
    predictor = p.StwfsapyPredictor(None, None, None, None, use_txt_vec=False)
    predictor.dfa_ = patched_dfa
    predictor.text_features_ = mk_text_features().fit([])
    in_txts = ["a", "bbb", "xx"]
    matches, ys = predictor.match_and_extend(
        in_txts,
        [[], [11, 14], [9]]
    )
    txt_indices = [
        idx
        for idx, txt
        in enumerate(in_txts)
        for _ in range(len(txt))
    ]
    for txt_idx, match, expected in zip(
            txt_indices,
            matches,
            expected_matches):
        check_fit_arg(
            None,
            predictor.text_features_,
            in_txts[txt_idx],
            match,
            expected
        )
    assert ys == [0, 0, 1, 0, 1, 0]


def test_match_and_extend_without_truth_without_vec(patched_dfa):
    predictor = p.StwfsapyPredictor(None, None, None, None, use_txt_vec=False)
    predictor.dfa_ = patched_dfa
    predictor.text_features_ = mk_text_features().fit([])
    in_txts = ["a", "bbb", "xx"]
    matches, counts = predictor.match_and_extend(in_txts)
    txt_indices = [
        idx
        for idx, txt
        in enumerate(in_txts)
        for _ in range(len(txt))
    ]
    for txt_idx, match, expected in zip(
            txt_indices,
            matches,
            expected_matches):
        check_fit_arg(
            None,
            predictor.text_features_,
            in_txts[txt_idx],
            match,
            expected
        )
    assert counts == [1, 3, 2]


def test_init_and_fit_without_vec(full_graph, mocker):
    predictor = p.StwfsapyPredictor(
        full_graph,
        c.test_type_concept,
        c.test_type_thesaurus,
        SKOS.broader,
        use_txt_vec=False)
    spy_deprecated = mocker.spy(t, "extract_deprecated")
    spy_case = mocker.spy(handlers, 'title_case_handler')
    predictor._init()
    assert predictor.text_vectorizer_ is None
    spy_deprecated.assert_called_once_with(full_graph)
    assert len(spy_case.mock_calls) == len(c.test_labels)
    for _, label in c.test_labels:
        assert call(label.toPython()) in spy_case.mock_calls
    assert isinstance(
        predictor.pipeline_.named_steps['Classifier'],
        DecisionTreeClassifier)
    combined = predictor.pipeline_.named_steps['Combined Features']
    assert isinstance(
        combined,
        ColumnTransformer)
    assert len(combined.transformers) == 4
    assert combined.transformers[0][0] == 'Thesaurus Features'
    assert combined.transformers[1][0] == 'Text Features'
    assert combined.transformers[2][0] == 'Position Features'
    assert combined.transformers[3][0] == 'Frequency Features'
    assert combined.transformers[0][1].thesaurus_relation == SKOS.broader
    spy_fit_pipeline = mocker.spy(predictor.pipeline_, "fit")
    predictor._fit_after_init(train_texts, y=train_labels)
    spy_fit_pipeline.assert_called_once()

    pipe_fit_arg_list = spy_fit_pipeline.call_args_list[0]
    assert len(pipe_fit_arg_list) == 2
    pipe_fit_args = pipe_fit_arg_list[0]
    assert pipe_fit_arg_list[1].get('y') == [1, 0, 1, 1]
    expected_features = [
            (c.test_concept_uri_0_0, [0], 1),
            (
                c.test_concept_uri_100_00,
                [4],
                1),
            (c.test_concept_uri_10_0, [0, 35], 0),
            (c.test_concept_uri_01_00, [17], 1),
            ]
    assert len(expected_features) == len(pipe_fit_args[0])
    txt_indices = [0, 3, 4, 4]
    for txt_idx, actual, expected in zip(
            txt_indices,
            pipe_fit_args[0],
            expected_features):
        check_fit_arg(
            None,
            predictor.text_features_,
            train_texts[txt_idx],
            actual,
            expected)


def check_fit_arg(vec_fun, text_feature_fun, txt, actual, expected):
    assert actual[0] == expected[0]
    assert (actual[1] == text_feature_fun.transform([txt])[0]).all()
    if vec_fun:
        assert (actual[2].toarray() ==
                vec_fun.transform([txt]).toarray()).all()
    else:
        assert actual[2] == 0
    assert actual[3] == len(txt)
    assert actual[4] == expected[1]
    assert actual[5] == expected[-1]


def test_predict(mocked_predictor):
    res = mocked_predictor.predict([])
    assert (
        res.toarray() == make_test_result_matrix(_classifications).toarray()
        ).all()
    mocked_predictor.match_and_extend.assert_called_once_with(
        []
    )
    mocked_predictor.pipeline_.predict.assert_called_once_with(
        _concepts_with_text
    )


def test_predict_proba(mocked_predictor):
    res = mocked_predictor.predict_proba([])
    assert (
        res.toarray() == make_test_result_matrix(
            _predictions[:, 1]).toarray()).all()
    mocked_predictor.match_and_extend.assert_called_once_with(
        []
    )
    mocked_predictor.pipeline_.predict_proba.assert_called_once_with(
        _concepts_with_text
    )


def test_suggest(mocked_predictor):
    res = mocked_predictor.suggest_proba([])
    assert res == [
        [(9, 0.2), (10, 0.3)],
        [(11, 0.4), (12, 0.5), (13, 0.6), (14, 0.7)],
        [(15, 0.8), (16, 0.9), (17, 1.0)],
        ]
    mocked_predictor.match_and_extend.assert_called_once_with(
        []
    )
    mocked_predictor.pipeline_.predict_proba.assert_called_once_with(
        _concepts_with_text
    )


def test_predict_no_match(no_match_predictor):
    res = no_match_predictor.predict([])
    assert res.getnnz() == 0
    assert res.shape == (3, len(_concept_map))


def test_predict_proba_no_match(no_match_predictor):
    res = no_match_predictor.predict_proba([])
    assert res.getnnz() == 0
    assert res.shape == (3, len(_concept_map))


def test_suggest_no_match(no_match_predictor):
    res = no_match_predictor.suggest_proba([])
    assert res == [[], [], []]


def test_fit(mocker):
    predictor = p.StwfsapyPredictor(None, None, None, None)
    predictor._init = mocker.Mock()
    predictor._fit_after_init = mocker.Mock()
    X = [list(range(i)) for i in range(13)]
    y = [i % 2 for i in range(13)]
    predictor.fit(X, y)
    predictor._init.assert_called_once()
    predictor._fit_after_init.assert_called_once_with(X, y=y)


def test_set_sentence_case(case_graph):
    predictor = p.StwfsapyPredictor(
        case_graph,
        c.test_type_concept,
        c.test_type_thesaurus,
        SKOS.broader,
        handle_title_case=False)
    predictor._init()
    assert 1 == len(list(predictor.dfa_.search("three word label")))
    assert 0 == len(list(predictor.dfa_.search("Three Word label")))


def test_set_title_case(case_graph):
    predictor = p.StwfsapyPredictor(
        case_graph,
        c.test_type_concept,
        c.test_type_thesaurus,
        SKOS.broader,
        handle_title_case=True)
    predictor._init()
    assert 1 == len(list(predictor.dfa_.search("three word label")))
    assert 1 == len(list(predictor.dfa_.search("Three Word label")))


def test_no_english_plural(case_graph):
    predictor = p.StwfsapyPredictor(
        case_graph,
        c.test_type_concept,
        c.test_type_thesaurus,
        SKOS.broader,
        simple_english_plural_rules=False,
    )
    predictor._init()
    assert 0 == len(list(predictor.dfa_.search('three word labels')))
    assert 1 == len(list(predictor.dfa_.search('three word label')))


def test_english_plural(case_graph):
    predictor = p.StwfsapyPredictor(
        case_graph,
        c.test_type_concept,
        c.test_type_thesaurus,
        SKOS.broader,
        simple_english_plural_rules=True,
    )
    predictor._init()
    assert 1 == len(list(predictor.dfa_.search('three word labels')))
    assert 1 == len(list(predictor.dfa_.search('three word labels')))


def test_english_plural_with_case(case_graph):
    predictor = p.StwfsapyPredictor(
        case_graph,
        c.test_type_concept,
        c.test_type_thesaurus,
        SKOS.broader,
        simple_english_plural_rules=True,
        handle_title_case=True
    )
    predictor._init()
    assert 1 == len(list(predictor.dfa_.search('three word labels')))
    assert 1 == len(list(predictor.dfa_.search('three word labels')))
    assert 1 == len(list(predictor.dfa_.search('Three Word Labels')))


def test_expansion(full_graph, mocker):
    stub = mocker.stub(name='expansion_stub')

    def mock_expander(**kwargs):
        return [stub]

    mocker.patch(
        'stwfsapy.expansion.collect_expansion_functions',
        mock_expander)
    predictor = p.StwfsapyPredictor(
        full_graph,
        c.test_type_concept,
        c.test_type_thesaurus,
        SKOS.broader,
        True,
        extract_upper_case_from_braces=False,
        extract_any_case_from_braces=True,
        expand_ampersand_with_spaces=False,
        expand_abbreviation_with_punctuation=False,
        simple_english_plural_rules=True)
    predictor._init()
    for _, label in c.test_labels:
        label_text = label.toPython()
        assert call(label_text) in stub.mock_calls


def test_mark_doc_end():
    predictor = p.StwfsapyPredictor(None, None, None, None)
    matches = [('a', None, None, [0], 5, 0), ('b', None, None, [2, 5], 7, 0)]
    predictor._mark_last_concept_in_doc(matches)
    assert matches == [
        ('a', None, None, [0], 5, 0),
        ('b', None, None, [2, 5], 7, 1)]


def test_mark_doc_end_empty():
    predictor = p.StwfsapyPredictor(None, None, None, None)
    lst = []
    predictor._mark_last_concept_in_doc(lst)
    assert lst == []


def test_uriref_str_inversion():
    ref = c.test_type_concept
    assert ref == p._load_uri_ref(p._store_uri_ref(ref))


def test_serialization_inversion_without_vec(tmpdir, full_graph):
    predictor = p.StwfsapyPredictor(
        full_graph,
        c.test_type_concept,
        c.test_type_thesaurus,
        SKOS.broader,
        use_txt_vec=False
    )
    predictor.fit(train_texts, train_labels)
    pth = tmpdir.mkdir("tmp").join("model.zip")
    predictor.store(pth.strpath)
    loaded = p.StwfsapyPredictor.load(pth.strpath)
    assert loaded.input == predictor.input
    assert not loaded.use_txt_vec
    assert loaded.extract_any_case_from_braces == (
        predictor.extract_any_case_from_braces)
    assert loaded.extract_upper_case_from_braces == (
        predictor.extract_upper_case_from_braces)
    assert loaded.expand_ampersand_with_spaces == (
        predictor.expand_ampersand_with_spaces)
    assert loaded.expand_abbreviation_with_punctuation == (
        predictor.expand_abbreviation_with_punctuation)
    assert loaded.simple_english_plural_rules == (
        predictor.simple_english_plural_rules)
    assert loaded.concept_type_uri == predictor.concept_type_uri
    assert loaded.sub_thesaurus_type_uri == predictor.sub_thesaurus_type_uri
    assert loaded.thesaurus_relation_type_uri == (
        predictor.thesaurus_relation_type_uri)
    assert loaded.thesaurus_relation_is_specialisation == (
        predictor.thesaurus_relation_is_specialisation)
    assert loaded.concept_map_ == predictor.concept_map_
    assert loaded.dfa_ == predictor.dfa_
    assert len(loaded.graph) == len(predictor.graph)
    assert loaded.concept_map_ == predictor.concept_map_
    assert loaded.dfa_ == predictor.dfa_
    assert len(loaded.graph) == len(predictor.graph)
    assert loaded.pipeline_[0].transformers[0][1].mapping_ == (
        predictor.pipeline_[0].transformers[0][1].mapping_)
    assert loaded.text_vectorizer_ is None
    loaded_txt_feat_names = [
        name
        for name, _
        in loaded.text_features_.transformer_list
    ]
    pred_txt_feat_names = [
        name
        for name, _
        in predictor.text_features_.transformer_list
    ]
    assert loaded_txt_feat_names == pred_txt_feat_names
    for triple in loaded.graph:
        assert triple in predictor.graph


def test_serialization_inversion_with_vec(tmpdir, full_graph):
    predictor = p.StwfsapyPredictor(
        full_graph,
        c.test_type_concept,
        c.test_type_thesaurus,
        SKOS.broader,
        use_txt_vec=True
    )
    predictor.fit(train_texts, train_labels)
    pth = tmpdir.mkdir("tmp").join("model.zip")
    predictor.store(pth.strpath)
    loaded = p.StwfsapyPredictor.load(pth.strpath)
    assert loaded.input == predictor.input
    assert loaded.use_txt_vec
    assert loaded.extract_any_case_from_braces == (
        predictor.extract_any_case_from_braces)
    assert loaded.extract_upper_case_from_braces == (
        predictor.extract_upper_case_from_braces)
    assert loaded.expand_ampersand_with_spaces == (
        predictor.expand_ampersand_with_spaces)
    assert loaded.expand_abbreviation_with_punctuation == (
        predictor.expand_abbreviation_with_punctuation)
    assert loaded.simple_english_plural_rules == (
        predictor.simple_english_plural_rules)
    assert loaded.concept_type_uri == predictor.concept_type_uri
    assert loaded.sub_thesaurus_type_uri == predictor.sub_thesaurus_type_uri
    assert loaded.thesaurus_relation_type_uri == (
        predictor.thesaurus_relation_type_uri)
    assert loaded.thesaurus_relation_is_specialisation == (
        predictor.thesaurus_relation_is_specialisation)
    assert loaded.concept_map_ == predictor.concept_map_
    assert loaded.dfa_ == predictor.dfa_
    assert len(loaded.graph) == len(predictor.graph)
    assert loaded.concept_map_ == predictor.concept_map_
    assert loaded.dfa_ == predictor.dfa_
    assert len(loaded.graph) == len(predictor.graph)
    assert loaded.pipeline_[0].transformers[0][1].mapping_ == (
        predictor.pipeline_[0].transformers[0][1].mapping_)
    assert loaded.text_vectorizer_ is not None
    assert loaded.text_vectorizer_.vocabulary_ == (
        predictor.text_vectorizer_.vocabulary_)
    loaded_txt_feat_names = [
        name
        for name, _
        in loaded.text_features_.transformer_list
    ]
    pred_txt_feat_names = [
        name
        for name, _
        in predictor.text_features_.transformer_list
    ]
    assert loaded_txt_feat_names == pred_txt_feat_names
    for triple in loaded.graph:
        assert triple in predictor.graph


def test_warning(mocker):
    logging_spy = mocker.spy(p, '_logger')
    con_state = ConstructionState(None, '', '')
    con_state.construct = mocker.Mock(side_effect=Exception())
    p._handle_construction(
        con_state,
        'test_concept',
        'invalid_label')
    logging_spy.warning.assert_called_once_with(
        'Could not process label "invalid_label" of concept "test_concept".'
    )



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from stwfsapy import text_features as tf
from sklearn.exceptions import NotFittedError
from stwfsapy.tests.upper_case_letters import upper_case_letters
import pytest

_text = "abcdefghijklmnopqrstuvwxyzäöü" + \
    "ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÜ" + \
    " " + \
    "0123456789" + \
    "\"'?!()&%$"


def test_count_char():
    assert tf._count_char(_text) == 78


def test_count_words():
    assert tf._count_word(_text) == 1


def test_count_special():
    assert tf._count_special(_text) == 6


def test_count_upper():
    assert tf._count_upper(_text) == 29


def test_count_digit():
    assert tf._count_digit(_text) == 10


def test_raises_when_not_fit():
    c_feature = tf.CountFeature(tf.CountType.N_CHAR)
    with pytest.raises(NotFittedError):
        c_feature.transform([])


def test_transform():
    c_feature = tf.CountFeature(tf.CountType.N_DIGIT)
    c_feature.fit([])
    transformed = c_feature.transform(["0123456789abc"[:i] for i in range(14)])
    assert len(transformed) == 14
    for i in range(10):
        assert list(transformed[i]) == [i]
    for i in range(10, 14):
        assert list(transformed[i]) == [10]


def test_config_char():
    c_feature = tf.CountFeature(tf.CountType.N_CHAR)
    c_feature.fit([])
    assert c_feature.function_ == tf._count_char


def test_config_word():
    c_feature = tf.CountFeature(tf.CountType.N_WORD)
    c_feature.fit([])
    assert c_feature.function_ == tf._count_word


def test_config_special():
    c_feature = tf.CountFeature(tf.CountType.N_SPECIAL)
    c_feature.fit([])
    assert c_feature.function_ == tf._count_special


def test_config_upper():
    c_feature = tf.CountFeature(tf.CountType.N_UPPER)
    c_feature.fit([])
    assert c_feature.function_ == tf._count_upper


def test_config_digit():
    c_feature = tf.CountFeature(tf.CountType.N_DIGIT)
    c_feature.fit([])
    assert c_feature.function_ == tf._count_digit


def test_feature_creation():
    union = tf.mk_text_features()
    assert [t[0] for t in union.transformer_list] == [
        tf._NAME_CHAR_FEATURE,
        tf._NAME_WORD_FEATURE,
        tf._NAME_SPECIAL_CHARS_FEATURE,
        tf._NAME_UPPER_FEATURE,
        tf._NAME_DIGIT_FEATURE,
    ]


def test_international_upper_case_recall():
    count = tf._count_upper(upper_case_letters)
    assert count == len(upper_case_letters)


def test_international_upper_case_precision():
    for c in upper_case_letters:
        assert 1 == tf._count_upper(f'xy{c}z')



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from stwfsapy.automata import nfa
from stwfsapy.automata import construction as const
from stwfsapy.automata import conversion as conv
from stwfsapy.tests.automata.data import accept


def test_alternation():
    graph = nfa.Nfa()
    construction = const.ConstructionState(graph, 'a|b|c', accept)
    construction.construct()
    graph.remove_empty_transitions()
    dfa_graph = conv.NfaToDfaConverter(graph).start_conversion()

    res = list(dfa_graph.search('a'))
    assert len(res) == 1
    assert res[0][0] == accept
    res = list(dfa_graph.search('b'))
    assert len(res) == 1
    assert res[0][0] == accept
    res = list(dfa_graph.search('c'))
    assert len(res) == 1
    assert res[0][0] == accept
    res = list(dfa_graph.search('d'))
    assert len(res) == 0


def test_alternation_center_kleene():
    graph = nfa.Nfa()
    construction = const.ConstructionState(graph, 'a|b*|c', accept)
    construction.construct()
    graph.remove_empty_transitions()
    dfa_graph = conv.NfaToDfaConverter(graph).start_conversion()

    res = list(dfa_graph.search(''))
    assert len(res) == 1
    assert res[0][0] == accept
    res = list(dfa_graph.search('bbbbb'))
    assert len(res) == 1
    assert res[0][0] == accept
    res = list(dfa_graph.search('a'))
    assert len(res) == 1
    assert res[0][0] == accept
    res = list(dfa_graph.search('c'))
    assert len(res) == 1
    assert res[0][0] == accept
    res = list(dfa_graph.search('aa'))
    assert len(res) == 0
    res = list(dfa_graph.search('cc'))
    assert len(res) == 0
    res = list(dfa_graph.search('ba'))
    assert len(res) == 0


def test_alternation_left_kleene():
    graph = nfa.Nfa()
    construction = const.ConstructionState(graph, 'a*|b|c', accept)
    construction.construct()
    graph.remove_empty_transitions()
    dfa_graph = conv.NfaToDfaConverter(graph).start_conversion()

    res = list(dfa_graph.search(''))
    assert len(res) == 1
    assert res[0][0] == accept
    res = list(dfa_graph.search('aaaaa'))
    assert len(res) == 1
    assert res[0][0] == accept
    res = list(dfa_graph.search('c'))
    assert len(res) == 1
    assert res[0][0] == accept
    res = list(dfa_graph.search('b'))
    assert len(res) == 1
    assert res[0][0] == accept
    res = list(dfa_graph.search('bb'))
    assert len(res) == 0
    res = list(dfa_graph.search('cc'))
    assert len(res) == 0
    res = list(dfa_graph.search('ac'))
    assert len(res) == 0


def test_alternation_right_kleene():
    graph = nfa.Nfa()
    construction = const.ConstructionState(graph, 'a|b|c*', accept)
    construction.construct()
    graph.remove_empty_transitions()
    dfa_graph = conv.NfaToDfaConverter(graph).start_conversion()

    res = list(dfa_graph.search(''))
    assert len(res) == 1
    assert res[0][0] == accept
    res = list(dfa_graph.search('ccccc'))
    assert len(res) == 1
    assert res[0][0] == accept
    res = list(dfa_graph.search('a'))
    assert len(res) == 1
    assert res[0][0] == accept
    res = list(dfa_graph.search('b'))
    assert len(res) == 1
    assert res[0][0] == accept
    res = list(dfa_graph.search('bb'))
    assert len(res) == 0
    res = list(dfa_graph.search('aa'))
    assert len(res) == 0
    res = list(dfa_graph.search('ca'))
    assert len(res) == 0



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from stwfsapy.util.passthrough_transformer import PassthroughTransformer
import numpy as np
import scipy.sparse as sp


def test_array_input():
    in_feat = [
        np.array([1, 0, 0]),
        np.array([0, 7, 0]),
        np.array([0, 0, -3])
    ]
    pt = PassthroughTransformer()
    out_feat = pt.transform(in_feat)
    assert type(out_feat) == np.ndarray
    assert (sp.diags([[1, 7, -3]], [0]).toarray() == out_feat).all()


def test_sparse_input():
    in_feat = [
        sp.lil_matrix(np.array([[1, 0, 0]])),
        sp.lil_matrix(np.array([[0, 7, 0]])),
        sp.lil_matrix(np.array([[0, 0, -3]]))
    ]
    pt = PassthroughTransformer()
    out_feat = pt.transform(in_feat)
    assert sp.issparse(out_feat)
    assert (sp.diags([[1, 7, -3]], [0]).toarray() == out_feat.toarray()).all()



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from stwfsapy import expansion as e
import common as c

replacement_fun_any = e._replace_by_pattern_fun(
    e._any_case_from_braces_expression)


def test_pattern_match():
    match = e._any_case_from_braces_expression.search(
        c.test_string_any
        )
    assert match.group(0) == "\\({}\\)".format(c.test_abbreviation_any)
    assert match.group(1) == c.test_abbreviation_any


def test_replacement():
    assert replacement_fun_any(c.test_string_any) == \
        c.test_abbreviation_any


def test_no_replacement():
    some_string = "some string\\)"
    assert replacement_fun_any(some_string) == some_string



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from rdflib.term import URIRef, Literal


test_ref_type = URIRef("http://zbw.eu/stw/descriptor")

_thsys_uri_print = "http://zbw.eu/stw/thsys/70265"
_thsys_uri_media = "https://zbw.eu/stw/thsys/181994"

thsys_ref_print = URIRef(_thsys_uri_print)
thsys_ref_media = URIRef(_thsys_uri_media)

thsys_prefLabel_print_en = Literal("Printed matters", lang="en")


_concept_uri_printed = "http://zbw.eu/stw/descriptor/14812-5"
_concept_uri_media = "http://zbw.eu/stw/descriptor/18211-4"

concept_ref_printed = URIRef(_concept_uri_printed)
concept_ref_media = URIRef(_concept_uri_media)

concept_prefLabel_printed_en = Literal("Printed Products", lang="en")
concept_prefLabel_printed_de = Literal("Druckerzeugnis", lang="de")
concept_prefLabel_printed_missing = Literal("Druckerzeugnis")

concept_altLabel_printed_en = Literal("Print Media", lang="en")

concept_prefLabel_media_en = Literal("Press media", lang="en")



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import stwfsapy.case_handlers as handlers

_inp = 'foo bar'
_inp_upper = 'F.?O.?O.?'


def test_uncase_first_char_lower():
    res = handlers.uncase_first_char('lower')
    assert res == '(L|l)ower'


def test_uncase_first_char_upper():
    res = handlers.uncase_first_char('Upper')
    assert res == '(U|u)pper'


def test_uncase_first_char_symbol():
    input = '\\?something'
    res = handlers.uncase_first_char(input)
    assert res == '\\?something'


def test_uncase_diacritic():
    input = 'ӟ'
    res = handlers.uncase_first_char(input)
    assert res == '(Ӟ|ӟ)'


def test_sentence_case_handles_upper():
    assert _inp_upper == handlers.sentence_case_handler(_inp_upper)


def test_title_case_handles_upper():
    assert _inp_upper == handlers.title_case_handler(_inp_upper)


def test_sentence_case_handler():
    res = handlers.sentence_case_handler(_inp)
    assert res == '(F|f)oo bar'


def test_title_case_handler():
    res = handlers.title_case_handler(_inp)
    assert res == '(F|f)oo (B|b)ar'



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from stwfsapy.util.set_closure import set_closure, RelationLoopException
from pytest import fixture, raises

_branching_k = 3
_depth = 5
_internal_nodes = sum(_branching_k**i for i in range(_depth))
_leafs = _branching_k ** _depth


@fixture
def tree_relation():
    return {
        i: {i * _branching_k + j for j in range(1, _branching_k+1)}
        for
        i
        in range(_internal_nodes)
    }


def _add_edge_to_tree_relation(relation, start, end, additionals=set()):
    relation[start].add(end)
    ret = additionals.copy()
    hlp_start = start
    # add all relations that should end up in the closure
    while hlp_start > 0:
        ret.add((hlp_start, end))
        hlp_start = (hlp_start-1) // _branching_k
    return relation, ret


@fixture
def single_diamond(tree_relation):
    diamond_top = _branching_k+1
    diamond_left = diamond_top*_branching_k+(_branching_k-1)
    diamond_right = diamond_top*_branching_k+_branching_k
    diamond_bottom = diamond_left*_branching_k+_branching_k
    return _add_edge_to_tree_relation(
        tree_relation,
        diamond_right,
        diamond_bottom)


@fixture
def double_diamond(single_diamond):
    tree_relation, additionals = single_diamond
    diamond_top = 1
    diamond_left = diamond_top*_branching_k+(_branching_k-1)
    diamond_right = diamond_top*_branching_k+_branching_k
    diamond_bottom = diamond_left*_branching_k+_branching_k
    tree_relation[diamond_right].add(diamond_bottom)
    additionals.add((diamond_right, diamond_bottom))
    return _add_edge_to_tree_relation(
        tree_relation,
        diamond_right,
        diamond_bottom,
        additionals)


@fixture
def double_diamond_reflexive(double_diamond):
    tree_relation, additionals = double_diamond
    for n in tree_relation:
        tree_relation, additionals = _add_edge_to_tree_relation(
            tree_relation,
            n,
            n,
            additionals
        )
    return tree_relation, additionals


def check_closure_edge(closures, start, end):
    hlp_start = start
    while hlp_start > 0:
        assert end in closures[hlp_start]
        hlp_start = (hlp_start-1) // _branching_k
    assert end in closures[0]


def check_tree_closure(closures, additional_relations={}):
    totals = _internal_nodes+_leafs
    assert len(closures) == totals
    for start in range(_internal_nodes):
        for end in range(1, _branching_k+1):
            check_closure_edge(closures, start, start*_branching_k + end)
    for start in range(totals):
        for end in closures[start]:
            ancestor = (end-1) // _branching_k
            assert ancestor == start or (
                ancestor in closures[start]) or (
                (start, end) in additional_relations) or (
                    end == start)


def test_closure_of_tree(tree_relation):
    closures = set_closure(tree_relation)
    check_tree_closure(closures)


def test_closure_diamond(single_diamond):
    p_order, additionals = single_diamond
    closures = set_closure(p_order)
    check_tree_closure(closures, additionals)


def test_closure_double_diamond(double_diamond):
    p_order, additionals = double_diamond
    closures = set_closure(p_order)
    check_tree_closure(closures, additionals)


def test_closure_double_diamond_reflexive(double_diamond_reflexive):
    p_order, additionals = double_diamond_reflexive
    closures = set_closure(p_order)
    check_tree_closure(closures, additionals)
    for n in p_order:
        assert n in closures[n]


def test_exception_on_cycle(tree_relation):
    tree_relation[_internal_nodes-2].add(_branching_k)
    with raises(RelationLoopException):
        set_closure(tree_relation)


def test_no_exception_on_non_cycle_backedge(tree_relation):
    tree_relation, additionals = _add_edge_to_tree_relation(
        tree_relation,
        _internal_nodes-2,
        1)
    closures = set_closure(tree_relation)
    check_tree_closure(closures, additionals)



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from stwfsapy.automata import nfa
from stwfsapy.automata import construction as c

from stwfsapy.tests.automata.data import accept

expression = 'test'


def test_init_with_empty_graph():
    graph = nfa.Nfa()
    construction = c.ConstructionState(graph, expression, accept)
    construction._set_up()
    assert len(graph.states) == 3
    assert graph.starts == []
    assert graph.states[0].non_word_char_transitions == {1}
    assert graph.states[1].empty_transitions == {2}
    assert construction.append_to == [2]
    assert construction.expression == expression
    assert construction.before_braces == [[1], [2]]
    assert construction.dangling_alternations.stack == []
    assert not construction.escape_next
    assert construction.accept == accept


def test_init_with_existing_graph(input_graph):
    construction = c.ConstructionState(input_graph, expression, accept)
    construction._set_up()
    assert len(input_graph.states) == 9
    assert input_graph.starts == [0, 1]
    assert input_graph.states[6].non_word_char_transitions == {7}
    assert input_graph.states[7].empty_transitions == {8}
    assert construction.append_to == [8]
    assert construction.expression == expression
    assert construction.before_braces == [[7], [8]]
    assert construction.dangling_alternations.stack == []
    assert not construction.escape_next
    assert construction.accept == accept


def test_adds_acceptance():
    graph = nfa.Nfa()
    construction = c.ConstructionState(graph, "a|b", accept)
    construction.construct()
    assert graph.starts == [0]
    assert graph.states[-1].accepts == [accept]
    assert graph.states[-2].accepts == [accept]
    assert graph.states[3].non_word_char_transitions == {7}
    assert graph.states[5].non_word_char_transitions == {6}
    assert graph.states[1].non_word_char_transitions == set()
    assert graph.states[1].empty_transitions == {2, 4}


def test_handles_multiple_alternations():
    graph = nfa.Nfa()
    construction = c.ConstructionState(graph, 'a|b|c', accept)
    construction.construct()
    assert graph.starts == [0]
    assert graph.states[-1].accepts == [accept]
    assert graph.states[-2].accepts == [accept]
    assert graph.states[-3].accepts == [accept]


def test_registers_escape():
    graph = nfa.Nfa()
    construction = c. ConstructionState(graph, '\\', accept)
    construction._set_up()
    construction._perform_step(0)
    assert graph.starts == []
    assert construction.escape_next


def test_handles_escaped_symbol(mocker):
    graph = nfa.Nfa()
    construction = c.ConstructionState(graph, '?', accept)
    construction._set_up()
    assert graph.starts == []
    construction.escape_next = True
    spy = mocker.spy(construction, "_process_symbol")
    construction._perform_step(0)
    spy.assert_called_once_with('?')


def test_handles_opening_brace(input_graph):
    append_to = [2, 4, 5]
    before_braces = [[0], [1, 3]]
    construction = c.ConstructionState(input_graph, '(', accept)
    construction._set_up()
    construction.append_to = append_to.copy()
    construction.before_braces = before_braces.copy()
    construction._perform_step(0)
    assert input_graph.starts == [0, 1]
    new_state_idx = len(input_graph.states) - 1
    assert construction.dangling_alternations.pop() == []
    assert construction.before_braces[:-2] == before_braces[:-1]
    assert construction.before_braces[-2] == append_to
    assert construction.before_braces[-1] == [new_state_idx]
    assert construction.append_to == [new_state_idx]
    for idx in append_to:
        assert new_state_idx in input_graph.states[idx].empty_transitions


def test_handles_closing_brace(input_graph, mocker):
    append_to = [2, 4, 5]
    old_append_len = len(append_to)
    before_braces = [[0], [1, 3]]
    ret = [12, 14]
    construction = c.ConstructionState(input_graph, ')', accept)
    construction._set_up()
    mocker.patch.object(
        construction.dangling_alternations,
        "pop",
        lambda: ret)
    construction.append_to = append_to.copy()
    construction.before_braces = before_braces.copy()
    construction._perform_step(0)
    assert input_graph.starts == [0, 1]
    assert construction.append_to[:old_append_len] == append_to
    assert construction.append_to[old_append_len:] == ret
    assert construction.before_braces == before_braces[:-1]


def test_handles_alternation(input_graph):
    append_to = [2, 4, 5]
    before_braces = [[0], [1, 3]]
    after_braces = [6, 7]
    construction = c.ConstructionState(input_graph, '|', accept)
    construction._set_up()
    construction.append_to = append_to.copy()
    construction.before_braces = before_braces.copy()
    construction.after_braces = after_braces.copy()
    construction._perform_step(0)
    assert input_graph.starts == [0, 1]
    assert construction.append_to == [len(construction.graph.states)-1]
    assert construction.dangling_alternations.stack[-1] == append_to


def test_handles_optional(input_graph):
    append_to = [2, 4, 5]
    before_braces = [[0], [1, 3]]
    after_braces = [6, 7]
    construction = c.ConstructionState(input_graph, '?', accept)
    construction._set_up()
    construction.append_to = append_to.copy()
    construction.before_braces = before_braces.copy()
    construction.after_braces = after_braces.copy()
    construction._perform_step(0)
    assert input_graph.starts == [0, 1]
    for idx in construction.append_to:
        for bb_idx in construction.before_braces[-1]:
            assert idx in input_graph.states[bb_idx].empty_transitions


def test_handles_kleene_closure(input_graph):
    append_to = [2, 4, 5]
    before_braces = [[0], [1, 3]]
    construction = c.ConstructionState(input_graph, '*', accept)
    construction._set_up()
    construction.append_to = append_to.copy()
    construction.before_braces = before_braces.copy()
    construction._perform_step(0)
    assert input_graph.starts == [0, 1]
    for idx in append_to:
        for bb_idx in construction.before_braces[-1]:
            assert bb_idx in input_graph.states[idx].empty_transitions
    for bb_idx in construction.before_braces[-1]:
        assert len(input_graph.states) - 1 in input_graph.states[
            bb_idx].empty_transitions



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" The tests in this file compare the behaviors of
stwfsapy and zaptain-stwfsa regarding overlap of potential matches."""
from stwfsapy.automata import nfa
import stwfsapy.automata.construction as const
import stwfsapy.automata.conversion as conv
import pytest


label_global = 'global'
id_global = 'id_global'
label_economic = 'economic'
id_economic = 'id_economic'
label_crisis = 'crisis'
id_crisis = 'id_crisis'
label_global_economic = 'global economic'
id_global_economic = 'id_global_economic'
label_economic_crisis = 'economic crisis'
id_economic_crisis = 'id_economic_crisis'
label_global_economic_crisis = 'global economic crisis'
id_global_economic_crisis = 'id_global_economic_crisis'


@pytest.fixture
def regression_test_graph():
    automaton = nfa.Nfa()
    concept_tuples = [
        (label_global, id_global),
        (label_economic, id_economic),
        (label_crisis, id_crisis),
        (label_global_economic, id_global_economic),
        (label_economic_crisis, id_economic_crisis),
        (label_global_economic_crisis, id_global_economic_crisis),
    ]
    for label, concept_id in concept_tuples:
        const.ConstructionState(automaton, label, concept_id).construct()
    automaton.remove_empty_transitions()
    dfa = conv.NfaToDfaConverter(automaton).start_conversion()
    return dfa


def test_regression_0(regression_test_graph):
    phrase = "global economic crisis unfolds"
    res = list(regression_test_graph.search(phrase))
    assert res == [
        (id_global_economic_crisis, label_global_economic_crisis, 0, 22)
    ]


def test_regression_1(regression_test_graph):
    phrase = "bank collapse threatens global economic system"
    res = list(regression_test_graph.search(phrase))
    assert res == [
        (id_global_economic, label_global_economic, 24, 39)
    ]


def test_regression_2(regression_test_graph):
    phrase = "regulatory bodies react to economic crisis"
    res = list(regression_test_graph.search(phrase))
    assert res == [
        (id_economic_crisis, label_economic_crisis, 27, 42)
    ]


def test_regression_3(regression_test_graph):
    phrase = "global trends in economic policy during the crisis"
    res = list(regression_test_graph.search(phrase))
    assert res == [
        (id_global, label_global, 0, 6),
        (id_economic, label_economic, 17, 25),
        (id_crisis, label_crisis, 44, 50),
    ]



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from stwfsapy.automata import nfa


symbol0 = 's'
symbol1 = 't'
accept = "Acceptance!"


@pytest.fixture
def input_graph():
    graph = nfa.Nfa()
    for _ in range(6):
        graph.add_state()
    graph.add_start(0)
    graph.add_start(1)
    graph.add_symbol_transition(0, 2, symbol0)
    graph.add_symbol_transition(1, 0, symbol0)
    graph.add_symbol_transition(0, 3, symbol1)
    graph.add_symbol_transition(1, 3, symbol1)
    graph.add_symbol_transition(2, 4, symbol0)
    graph.add_symbol_transition(4, 5, symbol1)
    graph.add_non_word_char_transition(1, 5)
    graph.add_acceptance(5, accept)
    return graph



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import defaultdict, OrderedDict
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.exceptions import NotFittedError
from math import log
import numpy as np


class FrequencyFeatures(BaseEstimator, TransformerMixin):

    def __init__(self):
        self.idfs_ = None
        self.log_doc_count_ = None

    def fit(self, X, y=None):
        concept_counts = defaultdict(int)
        concepts = []
        # count one doc more to cover case where a concept does not appear.
        doc_count = 1
        for x in X:
            concepts.append(x[0])
            if x[-1] == 1:
                for concept in concepts:
                    concept_counts[concept] += 1
                doc_count += 1
                concepts = []
        self.log_doc_count_ = log(doc_count)
        idfs = dict()
        for concept, count in concept_counts.items():
            idfs[concept] = log(doc_count/(count+1))
        self.idfs_ = idfs
        return self

    def transform(self, X, y=None):
        if self.idfs_ is None:
            raise NotFittedError
        ret = np.zeros((len(X), 3))
        concept_counts = OrderedDict()
        ret_ptr = 0
        doc_concept_sum = 0
        for x in X:
            concept = x[0]
            concept_count = len(x[1])
            concept_counts[concept] = concept_count
            doc_concept_sum += concept_count
            if x[-1] == 1:
                for concept, count in concept_counts.items():
                    tf = count/doc_concept_sum
                    idf = self.idfs_.get(concept, self.log_doc_count_)
                    ret[ret_ptr, 0] = tf
                    ret[ret_ptr, 1] = idf
                    ret[ret_ptr, 2] = tf*idf
                    ret_ptr += 1
                doc_concept_sum = 0
                concept_counts = OrderedDict()
        return ret



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from stwfsapy.automata import heap
import pytest


def check_heap(queue):
    for i in range(1, len(queue.heap)):
        assert queue.heap[i][0] >= queue.heap[heap._parent_index(i)][0]
    for key in queue.mapping:
        assert key == queue.heap[queue.mapping[key]][1]


@pytest.fixture
def some_heap():
    queue = heap.BinaryMinHeap()
    for i in range(12):
        key = 2 * i
        val = 3 * i
        queue.heap.append((key, val))
        queue.mapping[val] = i
    return queue


def test_sift_up(some_heap):
    check_heap(some_heap)
    idx = 7
    old_key, old_val = some_heap.heap[idx]
    some_heap.heap[idx] = (0, old_val)
    some_heap._sift_up(idx)
    check_heap(some_heap)
    assert some_heap.heap[0] == (0, 0)
    assert some_heap.heap[1] == (0, old_val)


def test_sift_down(some_heap):
    idx = 1
    old_key, old_val = some_heap.heap[idx]
    some_heap.heap[idx] = (17, old_val)
    some_heap._sift_down(idx)
    check_heap(some_heap)
    assert some_heap.heap[7] == (17, old_val)


def test_sift_down_already_correct(some_heap):
    idx = 1
    old_key, old_val = some_heap.heap[1]
    some_heap._sift_down(idx)
    check_heap(some_heap)
    assert some_heap.heap[1] == (old_key, old_val)


def test_parent_idx():
    for i in range(200):
        assert heap._parent_index(heap._lchild_index(i)) == i
        assert heap._parent_index(heap._rchild_index(i)) == i


def test_push_empty():
    queue = heap.BinaryMinHeap()
    queue.push(5, 0)
    assert len(queue.heap) == 1
    assert len(queue.mapping) == 1
    assert queue.heap[0] == (0, 5)
    assert queue.mapping[5] == 0


def test_push(some_heap, mocker):
    spy = mocker.spy(some_heap, "_sift_up")
    some_heap.push(15, 13)
    spy.assert_called_once_with(len(some_heap.heap)-1)


def test_pop_(some_heap, mocker):
    spy = mocker.spy(some_heap, "_sift_down")
    head = some_heap.pop()
    assert head == 0
    spy.assert_called_once_with(0)
    check_heap(some_heap)


def test_pop_empty():
    empty = heap.BinaryMinHeap()
    with pytest.raises(IndexError):
        empty.pop()


def test_can_increase_key(some_heap, mocker):
    spy_down = mocker.spy(some_heap, "_sift_down")
    spy_up = mocker.spy(some_heap, "_sift_up")
    top = some_heap.heap[0]
    some_heap.change_priority(top[1], 256)
    check_heap(some_heap)
    spy_down.assert_called_once_with(0)
    spy_up.assert_not_called()


def test_can_decrease_key(some_heap, mocker):
    spy_down = mocker.spy(some_heap, "_sift_down")
    spy_up = mocker.spy(some_heap, "_sift_up")
    last = some_heap.heap[-1]
    some_heap.change_priority(last[1], 0)
    check_heap(some_heap)
    spy_up.assert_called_once_with(len(some_heap.heap)-1)
    spy_down.assert_not_called()


def test_can_set_equal_key(some_heap, mocker):
    spy_down = mocker.spy(some_heap, "_sift_down")
    spy_up = mocker.spy(some_heap, "_sift_up")
    middle = some_heap.heap[5]
    some_heap.change_priority(middle[1], middle[0])
    spy_up.assert_not_called()
    spy_down.assert_not_called()



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from stwfsapy.util import input_handler as h
import pytest

file_content = "Some text inside a file"


@pytest.fixture
def handler_file(tmp_path):
    p = tmp_path / "hello.txt"
    p.write_text(file_content)
    return p


def test_content_handler():
    assert file_content == h.handle_content(file_content)


def test_file_handler(handler_file):
    with open(handler_file) as f:
        assert file_content == h.handle_file(f)


def test_filepath_handler(handler_file):
    assert file_content == h.handle_filename(handler_file.resolve())


def test_get_content_handler():
    assert h.handle_content == h.get_input_handler('content')


def test_get_file_handler():
    assert h.handle_file == h.get_input_handler('file')


def test_get_filename_handler():
    assert h.handle_filename == h.get_input_handler('filename')


def test_get_unknown_handler():
    with pytest.raises(h.UnknownInputTypeException):
        h.get_input_handler('dfsjdfs')



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from stwfsapy.expansion import simple_english_plural_fun


def test_trailing_non_alpha():
    label = 'fo0'
    assert simple_english_plural_fun(label) == label


def test_trailing_s():
    label = 'bas'
    assert simple_english_plural_fun(label) == label


def test_default_endingn():
    label = 'bar'
    assert simple_english_plural_fun(label) == 'bars?'


def test_trailing_y():
    label = 'daily'
    prefix_idx = len(label)-1
    res = simple_english_plural_fun(label)
    assert len(res) == len(label) + 6
    assert label[:prefix_idx] == res[:prefix_idx]
    assert res[prefix_idx:] == '(y|ies)'



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
import stwfsapy.thesaurus as t
import stwfsapy.tests.thesaurus.common as c


@pytest.fixture
def patch_module(mocker):
    thesaurus_module = t
    mocker.patch.object(
        thesaurus_module, "_filter_by_langs")
    mocker.patch.object(
        thesaurus_module, "filter_subject_tuples_from_set")
    mocker.patch.object(
        thesaurus_module, "_unwrap_labels")
    # return mocker


@pytest.fixture
def concept_set():
    return {
        c.concept_ref_printed,
        c.concept_ref_media
    }


def test_no_language_option(label_graph, patch_module):
    t.retrieve_concept_labels(label_graph, langs=set())
    t._filter_by_langs.assert_not_called()


def test_none_language_option(label_graph, patch_module):
    t.retrieve_concept_labels(label_graph, langs=None)
    t._filter_by_langs.assert_not_called()


def test_language_option(label_graph, patch_module):
    t.retrieve_concept_labels(label_graph, langs={"en"})
    t._filter_by_langs.assert_called()


def test_no_prefix_option(label_graph, patch_module):
    t.retrieve_concept_labels(label_graph)
    t.filter_subject_tuples_from_set.assert_not_called()


def test_none_prefix_option(label_graph, patch_module):
    t.retrieve_concept_labels(label_graph, allowed=None)
    t.filter_subject_tuples_from_set.assert_not_called()


def test_prefix_option(label_graph, concept_set, patch_module):
    t.retrieve_concept_labels(
        label_graph,
        concept_set
    )
    t.filter_subject_tuples_from_set.assert_called()


def test_integration(typed_label_graph, concept_set):
    result = list(t.retrieve_concept_labels(
        typed_label_graph,
        allowed=concept_set,
        langs={"en"},
    ))
    assert (
        c.concept_ref_printed,
        c.concept_prefLabel_printed_en.value) in result
    assert (
        c.concept_ref_printed,
        c.concept_altLabel_printed_en.value) in result
    assert (
        c.concept_ref_printed,
        c.concept_prefLabel_printed_de.value) not in result
    assert c.thsys_ref_print not in map(lambda t: t[0], result)



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from numpy import array
from rdflib.namespace import SKOS
from stwfsapy import thesaurus as t
from stwfsapy import thesaurus_features as tf
from stwfsapy.tests.thesaurus import common as tc
from stwfsapy.tests import common as c
from scipy.sparse import coo_matrix, csr_matrix
from sklearn.exceptions import NotFittedError
import pytest


def test_collect_po_from_tuples():
    tuples = [
        (tc.concept_ref_printed, tc.concept_ref_media),
        (tc.concept_ref_media, tc.thsys_ref_media),
        (tc.concept_ref_printed, tc.thsys_ref_print)
    ]
    po = tf._collect_po_from_tuples(tuples)
    assert po == {
        tc.concept_ref_printed: {tc.concept_ref_media, tc.thsys_ref_print},
        tc.concept_ref_media: {tc.thsys_ref_media}
    }


def test_unfitted_raises():
    feat = tf.ThesaurusFeatureTransformation(None, None, None, None)
    with pytest.raises(NotFittedError):
        feat.transform([])


def test_transform():
    trans = tf.ThesaurusFeatureTransformation(None, None, None, None)
    trans.mapping_ = {
        'a': coo_matrix([[1]]), 'b': coo_matrix([[2]]), 'c': coo_matrix([[3]])}
    res = trans.transform(['c', 'c', 'a'])
    assert (res.toarray() == array([[3], [3], [1]])).all()


def test_fit(full_graph):
    concepts = set(t.extract_by_type_uri(
        full_graph,
        c.test_type_concept))
    thesauri = set(t.extract_by_type_uri(
        full_graph,
        c.test_type_thesaurus))
    trans = tf.ThesaurusFeatureTransformation(
        full_graph,
        concepts,
        thesauri,
        SKOS.broader)
    trans.fit()
    mapping = trans.mapping_
    assert len(mapping) == len(c.test_concepts)
    for x in mapping.values():
        assert x.shape[1] == 6
    # Can not test positions because retrieval from graph is not deterministic.
    # Therefore, test non zero entries only.
    assert mapping[c.test_concept_uri_0_0].getnnz() == 1
    assert mapping[c.test_concept_uri_01_0].getnnz() == 2
    assert mapping[c.test_concept_uri_01_00].getnnz() == 2
    assert mapping[c.test_concept_uri_10_0].getnnz() == 2
    assert mapping[c.test_concept_uri_10_1].getnnz() == 2
    assert mapping[c.test_concept_uri_100_0].getnnz() == 3
    assert mapping[c.test_concept_uri_100_00].getnnz() == 3
    assert mapping[c.test_concept_uri_100_01].getnnz() == 3
    assert mapping[c.test_concept_uri_100_02].getnnz() == 3


def test_transform_unknown():
    trans = tf.ThesaurusFeatureTransformation(
        None,
        None,
        None,
        None,)

    feature_dim = 12
    trans.feature_dim_ = feature_dim
    known = csr_matrix(([1], ([0], [4])), shape=(1, feature_dim))
    trans.mapping_ = {'key': known}
    random_results = trans.transform([
        'some random stuff edsfysdfhjsedf',
        'key'])
    assert random_results.shape == (2, feature_dim)
    assert random_results.getrow(0).getnnz() == 0
    assert random_results.getrow(1).getnnz() == 1


def test_empty_relation(full_graph):
    trans = tf.ThesaurusFeatureTransformation(
        full_graph,
        set(t.extract_by_type_uri(full_graph, c.test_type_concept)),
        set(),
        "",
    )
    trans.fit([], [])
    features = trans.transform(['empty'])
    assert features.shape == (1, 1)
    assert features.getnnz() == 0



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from stwfsapy import thesaurus as t
from stwfsapy.tests.thesaurus import common as c


def test_extract_by_type_uri(typed_label_graph):
    res = list(t.extract_by_type_uri(typed_label_graph, c.test_ref_type))
    assert len(res) == 2
    assert c.concept_ref_printed in res
    assert c.concept_ref_media in res


def test_extract_by_type_uri_with_remove(typed_label_graph):
    res = list(t.extract_by_type_uri(
        typed_label_graph,
        c.test_ref_type,
        {c.concept_ref_printed}))
    assert res == [c.concept_ref_media]



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import scipy.sparse as sp
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin


class PassthroughTransformer(BaseEstimator, TransformerMixin):
    ''' Helper Class to better handle array input for ColumnTransformer.'''

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        if sp.issparse(X[0]):
            ret = sp.vstack(X, format='csr')
        else:
            ret = np.vstack(X)
        return ret



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from stwfsapy.position_features import PositionFeatures
import numpy as np


position_feature_data = [
    (3, [3, 4, 0, 2]),
    (6, [1, 8, 23, 12]),
    (15, [12, 36, 25, 7]),
    (12, [8, 102, 17, 9, 20]),
    (70, [13]),
]


def test_handle_empty():
    features = PositionFeatures()
    assert features.transform([]).tolist() == np.zeros((0, 3)).tolist()


def test_convert_docs():
    features = PositionFeatures()
    res = features.transform(position_feature_data)
    assert res.tolist() == [
        [0/3, 4/3, 4/3],
        [1/6, 23/6, 22/6],
        [7/15, 36/15, 29/15],
        [8/12, 102/12, 94/12],
        [13/70, 13/70, 0]]



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from sklearn.base import BaseEstimator, TransformerMixin
import numpy as np


class PositionFeatures(BaseEstimator, TransformerMixin):

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        out = np.zeros((len(X), 3))
        for idx, x in enumerate(X):
            positions = x[1]
            min_pos = min(positions)
            max_pos = max(positions)
            spread = max_pos - min_pos
            txt_len = x[0]
            out[idx][0] = min_pos / txt_len
            out[idx][1] = max_pos / txt_len
            out[idx][2] = spread / txt_len
        return out



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def uncase_first_char(text: str) -> str:
    if text[0].isalpha():
        return "({}|{}){}".format(
            text[0].upper(),
            text[0].lower(),
            text[1:]
            )
    else:
        return text


def sentence_case_handler(text: str) -> str:
    if text.isupper():
        return text
    return uncase_first_char(text)


def title_case_handler(text: str) -> str:
    if text.isupper():
        return text
    return ' '.join(map(uncase_first_char, text.split()))



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rdflib.namespace import OWL
from rdflib.term import Literal
from stwfsapy import thesaurus as t
from stwfsapy.tests.thesaurus import common as c


def test_extract_deprecated(label_graph):
    label_graph.add((c.concept_ref_printed, OWL.deprecated, Literal(True)))
    res = list(t.extract_deprecated(label_graph))
    assert res == [c.concept_ref_printed]



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import stwfsapy.expansion as e

first = 'R'
second = 'D'
ampersand_string = "{}&{}".format(first, second)
multi_ampersand_string = "A&" + ampersand_string


def test_matches():
    match = e._ampersand_abbreviation_matcher.search(ampersand_string)
    assert match is not None
    assert match.group(0) == ampersand_string
    assert match.group(1) == first
    assert match.group(2) == second


def test_no_match_at_end_of_string():
    match = e._ampersand_abbreviation_matcher.search(" "+ampersand_string)
    assert match is None


def test_no_match_at_start_of_string():
    match = e._ampersand_abbreviation_matcher.search(ampersand_string + " ")
    assert match is None


def test_no_match_multiple_ampersand():
    match = e._ampersand_abbreviation_matcher.search(multi_ampersand_string)
    assert match is None


def test_replacement():
    replaced = e._expand_ampersand_with_spaces_fun(ampersand_string)
    assert len(replaced) == len(ampersand_string) + 4
    for i in range(len(ampersand_string)-1):
        offset = i * 3
        assert replaced[offset] == ampersand_string[i]
        assert replaced[offset+1] == ' '
        assert replaced[offset+2] == '?'
    assert replaced[-1] == ampersand_string[-1]


def test_no_replacement_multiple_ampersand():
    replaced = e._expand_ampersand_with_spaces_fun(multi_ampersand_string)
    assert replaced == multi_ampersand_string



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from stwfsapy.automata import dfa

symbol = 's'


@pytest.fixture
def two_state_graph():
    graph = dfa.Dfa()
    graph.add_state()
    graph.add_state()
    return graph


@pytest.fixture
def foo_graph(two_state_graph):
    two_state_graph.set_non_word_char_transition(0, 1)
    two_state_graph.add_state()
    two_state_graph.set_symbol_transition(1, 2, 'f')
    two_state_graph.add_state()
    two_state_graph.set_symbol_transition(2, 3, 'o')
    two_state_graph.set_symbol_transition(3, 3, 'o')
    two_state_graph.add_state()
    two_state_graph.set_non_word_char_transition(3, 4)
    two_state_graph.add_acceptances(4, ["bar"])
    return two_state_graph


def test_add_state(two_state_graph):
    old_len = len(two_state_graph.states)
    two_state_graph.add_state()
    assert len(two_state_graph.states) == old_len + 1


def test_add_acceptance(two_state_graph):
    accept = "Acceptance!"
    idx = 1
    two_state_graph.add_acceptances(idx, [accept])
    assert accept in two_state_graph.states[idx].accepts


def test_set_symbol_transition(two_state_graph):
    two_state_graph.set_symbol_transition(0, 1, symbol)
    assert two_state_graph.states[0].symbol_transitions[symbol] == 1


def test_override_symbol_transition(two_state_graph):
    two_state_graph.set_symbol_transition(0, 1, symbol)
    idx = two_state_graph.add_state()
    two_state_graph.set_symbol_transition(0, idx, symbol)
    assert two_state_graph.states[0].symbol_transitions[symbol] == idx


def test_set_non_word_char_transition(two_state_graph):
    two_state_graph.set_non_word_char_transition(0, 1)
    assert two_state_graph.states[0].non_word_char_transition == 1


def test_search(foo_graph):
    text = "fooo"
    res = list(foo_graph.search(text))
    assert len(res) == 1
    assert res[0][0] == 'bar'
    assert res[0][1] == text
    assert res[0][2] == 0
    assert res[0][3] == 4


def test_serialization_inversion(foo_graph):
    assert foo_graph == dfa.Dfa.from_dict(
        foo_graph.to_dict(lambda x: x), lambda x: x)



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import stwfsapy.thesaurus as t


def test_label_unwrap(tuples):
    unwrapped = list(t._unwrap_labels(tuples))
    assert len(tuples) == len(unwrapped)
    for original, transformed in zip(tuples, unwrapped):
        assert isinstance(transformed[1], str)
        assert original[1].toPython() == transformed[1]
        assert original[0] == transformed[0]



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pytest import fixture
from rdflib.graph import Graph
from rdflib.namespace import RDF, SKOS
from stwfsapy.tests import common as c


@fixture
def full_graph():
    g = Graph()
    for t in c.test_thesauri:
        g.add((t, RDF.type, c.test_type_thesaurus))
    for cncpt in c.test_concepts:
        g.add((cncpt, RDF.type, c.test_type_concept))
    for narrow, broader in c.thsrs_broader:
        g.add((narrow, SKOS.broader, broader))
    for narrow, broader in c.cncpt_broaders:
        g.add((narrow, SKOS.broader, broader))
    for cncpt, label in c.test_labels:
        g.add((cncpt, SKOS.prefLabel, label))
    return g



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from stwfsapy.tests.thesaurus import common as c
from rdflib import Graph
from rdflib.namespace import SKOS, RDF


@pytest.fixture
def tuples():
    return [
        (c.concept_ref_printed, c.concept_prefLabel_printed_en),
        (c.concept_ref_media, c.concept_prefLabel_media_en),
        (c.concept_ref_printed, c.concept_prefLabel_printed_missing)
    ]


@pytest.fixture
def label_graph():
    g = Graph()
    g.add((
        c.concept_ref_printed,
        SKOS.prefLabel,
        c.concept_prefLabel_printed_en))
    g.add((
        c.concept_ref_printed,
        SKOS.altLabel,
        c.concept_altLabel_printed_en))
    g.add((
        c.concept_ref_media,
        SKOS.altLabel,
        c.concept_altLabel_printed_en))
    return g


@pytest.fixture
def typed_label_graph(label_graph):
    g = label_graph
    g.add((
        c.concept_ref_printed,
        SKOS.prefLabel,
        c.concept_prefLabel_printed_de))
    g.add((
        c.thsys_ref_print,
        SKOS.prefLabel,
        c.thsys_prefLabel_print_en))
    g.add((
        c.concept_ref_media,
        RDF.type,
        c.test_ref_type))
    g.add((
        c.concept_ref_printed,
        RDF.type,
        c.test_ref_type))
    return g



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Callable, Pattern, List
import re


_symbol_base_expression = re.compile(r'([\[\]()\{\}*?])')

_upper_case_abbreviation_from_braces_expression = re.compile(
   r"^([A-Z-]{2,})\s*\\\(.{3,}?\\\)"
)

_any_case_from_braces_expression = re.compile(
    r"\\\((.{3,}?)\\\)"
)

_ampersand_abbreviation_matcher = re.compile(r"^([A-Z])&([A-Z])$")
"""Matches terms like R&D.
Assumes that there are not multiple ampersands and only single
capital letters at each side."""


_upper_case_matcher = re.compile(r"([A-Z])")
"""Matches capital letters.
This is used for inserting optional dots between the letters."""


def base_expansion(input: str):
    return _symbol_base_expression.sub(r'\\\g<1>', input)


def _replace_by_pattern_fun(pattern: Pattern[str]) -> Callable[[str], str]:
    def replacer(label: str) -> str:
        match = pattern.search(label)
        if match is None:
            return label
        return match[1]
    return replacer


def _expand_ampersand_with_spaces_fun(label: str) -> str:
    return _ampersand_abbreviation_matcher.sub(
        r"\g<1> ?& ?\g<2>", label)


def _expand_abbreviation_with_punctuation_fun(label: str) -> str:
    if label.isupper() and label.isalpha():
        return _upper_case_matcher.sub(r"\g<0>\\.?", label)
    return label


def simple_english_plural_fun(label: str):
    last = label[-1]
    if last == 'y':
        return label[:-1]+'(y|ies)'
    if last.isalpha() and not (last == 's' or last == 'S'):
        return label + 's?'
    return label


def collect_expansion_functions(
        extract_upper_case_from_braces: bool = True,
        extract_any_case_from_braces: bool = False,
        expand_ampersand_with_spaces: bool = True,
        expand_abbreviation_with_punctuation: bool = True,
        ) -> List[Callable[[str], str]]:
    options = (
        (
            extract_upper_case_from_braces,
            _replace_by_pattern_fun(
                _upper_case_abbreviation_from_braces_expression)
        ),
        (
            extract_any_case_from_braces,
            _replace_by_pattern_fun(_any_case_from_braces_expression)
        ),
        (
            expand_ampersand_with_spaces,
            _expand_ampersand_with_spaces_fun
        ),
        (
            expand_abbreviation_with_punctuation,
            _expand_abbreviation_with_punctuation_fun
        ),
    )
    ret = [fun for flag, fun in options if flag]
    ret.insert(0, base_expansion)
    return ret



# Copyright 2020-2023 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from stwfsapy.util.passthrough_transformer import PassthroughTransformer
from stwfsapy.frequency_features import FrequencyFeatures
from stwfsapy.position_features import PositionFeatures
from collections import defaultdict
from typing import Dict, FrozenSet, List, Iterable, \
    Container, Tuple, TypeVar, Union
from scipy.sparse import spmatrix
from numpy import array
from logging import getLogger
from rdflib.term import URIRef
from rdflib import Graph
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.pipeline import Pipeline
from sklearn.compose import ColumnTransformer
from sklearn.tree import DecisionTreeClassifier
from sklearn.feature_extraction.text import TfidfVectorizer
from scipy.sparse import csr_matrix
from stwfsapy import thesaurus as t
from stwfsapy.automata import nfa, construction, conversion, dfa
from stwfsapy.thesaurus_features import ThesaurusFeatureTransformation
from stwfsapy.text_features import mk_text_features
from stwfsapy.util.input_handler import get_input_handler
from stwfsapy import case_handlers
from stwfsapy import expansion
import pickle as pkl
from json import dumps, loads
from zipfile import ZipFile


T = TypeVar('T')
N = TypeVar('N', int, float)
Nl = TypeVar('Nl', List[int], List[float])


_KEY_DFA = 'dfa'
_KEY_CONCEPT_MAP = 'concept_map'
_KEY_CONCEPT_TYPE_URI = 'concept_type_uri'
_KEY_THESAURUS_TYPE_URI = 'thesaurus_type_uri'
_KEY_THESAURUS_RELATION_TYPE_URI = 'thesaurus_relation_type_uri'
_KEY_THESAURUS_RELATION_IS_SPECIALISATION = (
    'thesaurus_relation_is_specialisation')
_KEY_REMOVE_DEPRECATED = 'remove_deprecated'
_KEY_LANGS = 'langs'
_KEY_INPUT = 'input'
_KEY_USE_TXT_VEC = 'use_txt_vec'
_KEY_HANDLE_TITLE_CASE = 'handle_title_case'
_KEY_EXTRACT_UPPER_CASE_FROM_BRACES = 'extract_upper_case_from_braces'
_KEY_EXTRACT_ANY_CASE_FROM_BRACES = 'extract_any_key_from_braces'
_KEY_EXPAND_AMPERSAND_WITH_SPACES = 'expand_ampersand_with_spaces'
_KEY_EXPAND_ABBREVIATION_WITH_PUNCTUATION = (
    'expand_abbreviation_with_punctuation')
_KEY_SIMPLE_ENGLISH_PLURAL_RULES = 'simple_english_plural_rules'

_NAME_GRAPH_FILE = 'graph.rdf'
_NAME_PIPELINE_FILE = 'pipeline.pkl'
_NAME_PREDICTOR_FILE = 'predictor.json'
_NAME_TEXT_FEATURES_FILE = 'text_features.pkl'
_NAME_TEXT_VECTORIZER_FILE = 'text_vectorizer.pkl'

_logger = getLogger('stwfsa')


class StwfsapyPredictor(BaseEstimator, ClassifierMixin):
    """Finds labels of thesaurus concepts in texts
    and assigns them a score."""
    def __init__(
            self,
            graph: Graph,
            concept_type_uri: Union[str, URIRef],
            sub_thesaurus_type_uri: Union[str, URIRef] = "",
            thesaurus_relation_type_uri: Union[str, URIRef] = "",
            thesaurus_relation_is_specialisation: bool = False,
            remove_deprecated: bool = True,
            langs: FrozenSet[str] = frozenset(),
            input: str = 'content',
            use_txt_vec: bool = False,
            handle_title_case: bool = True,
            extract_upper_case_from_braces: bool = True,
            extract_any_case_from_braces: bool = False,
            expand_ampersand_with_spaces: bool = True,
            expand_abbreviation_with_punctuation: bool = True,
            simple_english_plural_rules: bool = False,
            ):
        """Creates the predictor.

        :param graph: The SKOS onthology used to extract the labels.
        :param concept_type_uri:
            The uri of the concept type.
            It is assumed that for every concept c,
            there is a triple (c, RDF.type, concept_type_uri)
            in the graph.
        :param  sub_thesaurus_type_uri: The uri of the concept type.
            It is assumed that for every sub thesaurus t,
            there is a triple (t, RDF.type, sub_thesaurus_type_uri)
            in the graph.
        :param  thesaurus_relation_type_uri:
            Uri of the relation that links concepts to thesauri.
        :param  thesaurus_relation_is_specialisation:
            Indicates whether the thesaurus_relation links thesauri to
            concepts or the other way round.
            E.g., for the relation skos:broader it should be false.
            Conversely it should be true for skos:narrower.
        :param  remove_deprecated:
            When True will discard deprecated subjects.
            Deprecation of a subject has to be indicated by
            a triple (s, OWL.deprecated, Literal(True)) in the graph.
        :param  langs:
            For each language present in the set,
            labels will be extracted from the graph.
            An empy set or None will extract labels regardless of language.
        :param  input:
            What type of input is presented to the fit method:

                * 'content': Input is expected to be an arraylike of string.
                * 'filename': Input is expected to be a list of filenames.
                * 'file': input is expected to be a list of file objects.
        :param  use_txt_vec:
            Whether to use vectorized representations of inputs.
            This can lead to high memory consumption.
        :param  handle_title_case:
            When True, will also match labels in title case.
            I.e., in a text the first letter of every word can be upper
            or lower case and will still be matched.
            When False only the case of the first word's first letter
            will be adapted.
            Example:

                  * Given a label "garbage can" and the
                    title "Oscar Lives in a Garbage Can"
                  * When handle_title_case == True
                    the label will match the text.
                  * When handle_title_case == False
                    the label will not match the text.
                    It would however still match
                    "Garbage can is home to grouchy neighbor.".
        :param extract_upper_case_from_braces:
            Removes the explanation in braces from labels.
            I.e., GDP (Gross Domestic Product) will be transformed to GDP.
        :param extract_any_case_from_braces:
            Can extract content of braces in labels.
            I.e., R&D (research and discovery) will be transformed to
            research and discovery.
            In contrast to extract_upper_case_from_braces it will extract
            the part inside the parenthesis and not the part before.
        :param expand_ampersand_with_spaces:
              For labels that contain an ampersand it will also match text
              containing spaces around that symbol.
              I.e., R & D will be matched for label R&D.
        :param expand_abbreviation_with_punctuation:
          For labels containing only uppercase letters it will also
          match text with punctuation added. I.e., G.D.P. for label GDP.
        :param simple_english_plural_rules:
          Can detect simple English plural forms of labels.
        """
        self.graph = graph
        if isinstance(concept_type_uri, str):
            concept_type_uri = URIRef(concept_type_uri)
        self.concept_type_uri = concept_type_uri
        if isinstance(sub_thesaurus_type_uri, str):
            sub_thesaurus_type_uri = URIRef(sub_thesaurus_type_uri)
        self.sub_thesaurus_type_uri = sub_thesaurus_type_uri
        if isinstance(thesaurus_relation_type_uri, str):
            thesaurus_relation_type_uri = URIRef(thesaurus_relation_type_uri)
        self.thesaurus_relation_type_uri = thesaurus_relation_type_uri
        self.thesaurus_relation_is_specialisation = (
            thesaurus_relation_is_specialisation)
        self.remove_deprecated = remove_deprecated
        self.langs = langs
        self.input = input
        self.use_txt_vec = use_txt_vec
        self.handle_title_case = handle_title_case
        self.extract_upper_case_from_braces = extract_upper_case_from_braces
        self.extract_any_case_from_braces = extract_any_case_from_braces
        self.expand_ampersand_with_spaces = expand_ampersand_with_spaces
        self.expand_abbreviation_with_punctuation = \
            expand_abbreviation_with_punctuation
        self.simple_english_plural_rules = simple_english_plural_rules

    def _init(self):
        all_deprecated = set(t.extract_deprecated(self.graph))
        concepts = set(t.extract_by_type_uri(
            self.graph,
            self.concept_type_uri,
            remove=all_deprecated))
        thesauri = set(t.extract_by_type_uri(
            self.graph,
            self.sub_thesaurus_type_uri,
            remove=all_deprecated
        ))
        self.concept_map_ = dict(zip(map(str, concepts), range(len(concepts))))
        thesaurus_features = ThesaurusFeatureTransformation(
            self.graph,
            concepts,
            thesauri,
            self.thesaurus_relation_type_uri,
            self.thesaurus_relation_is_specialisation
        )
        labels = t.retrieve_concept_labels(
            self.graph,
            allowed=concepts,
            langs=self.langs)
        nfautomat = nfa.Nfa()
        if self.handle_title_case:
            case_handler = case_handlers.title_case_handler
        else:
            case_handler = case_handlers.sentence_case_handler
        expansion_funs = expansion.collect_expansion_functions(
            extract_upper_case_from_braces=self.extract_upper_case_from_braces,
            extract_any_case_from_braces=self.extract_any_case_from_braces,
            expand_ampersand_with_spaces=self.expand_ampersand_with_spaces,
            expand_abbreviation_with_punctuation=(
                self.expand_abbreviation_with_punctuation),
        )
        if self.simple_english_plural_rules:
            plural_fun = expansion.simple_english_plural_fun
        else:
            def plural_fun(x):
                return x
        for concept, label in labels:
            expanded = label
            for f in expansion_funs:
                expanded = f(expanded)
            _handle_construction(
                construction.ConstructionState(
                    nfautomat,
                    plural_fun(case_handler(expanded)),
                    str(concept)
                ),
                concept,
                label)
        nfautomat.remove_empty_transitions()
        converter = conversion.NfaToDfaConverter(nfautomat)
        self.dfa_ = converter.start_conversion()
        self.text_features_ = mk_text_features().fit([])
        transformations = [
                    ("Thesaurus Features", thesaurus_features, 0),
                    ('Text Features', PassthroughTransformer(), 1),
                    ('Position Features', PositionFeatures(), [3, 4]),
                    (
                        'Frequency Features',
                        FrequencyFeatures(),
                        [0, 4, 5])
                ]
        if self.use_txt_vec:
            self.text_vectorizer_ = TfidfVectorizer(input=self.input)
            transformations.append(
                ('Text Vector', PassthroughTransformer(), 2),
            )
        else:
            self.text_vectorizer_ = None
        self.pipeline_ = Pipeline([
            ("Combined Features", ColumnTransformer(
                transformations)),
            ("Classifier", DecisionTreeClassifier(
                min_samples_leaf=25,
                max_leaf_nodes=100))
        ])

    def fit(self, X, y=None, **kwargs):
        self._init()
        return self._fit_after_init(X, y=y)

    def _fit_after_init(self, X, y=None):
        if self.use_txt_vec:
            self.text_vectorizer_.fit(X)
        matches, train_y = self.match_and_extend(X, y)
        self.pipeline_.fit(matches, y=train_y)
        return self

    def predict_proba(self, X) -> csr_matrix:
        match_X, doc_counts = self.match_and_extend(X)
        if match_X:
            predictions = self.pipeline_.predict_proba(match_X)[:, 1]
        else:
            predictions = []
        return self._create_sparse_matrix(
            predictions,
            [tpl[0] for tpl in match_X],
            doc_counts
        )

    def suggest_proba(
            self,
            texts
            ) -> List[List[Tuple[str, float]]]:
        """For a given list of texts,
        this method returns the matched concepts and their scores."""
        match_X, doc_counts = self.match_and_extend(texts)
        if match_X:
            predictions = self.pipeline_.predict_proba(match_X)[:, 1]
        else:
            predictions = []
        combined = StwfsapyPredictor._collect_prediction_results(
            predictions,
            [tpl[0] for tpl in match_X],
            doc_counts
        )
        return [
            [
                (concept, score)
                for concept, score
                in zip(concepts, scores)
            ]
            for concepts, scores
            in combined
        ]

    def predict(self, X) -> csr_matrix:
        match_X, doc_counts = self.match_and_extend(X)
        if match_X:
            predictions = self.pipeline_.predict(match_X)
        else:
            predictions = []
        return self._create_sparse_matrix(
            predictions,
            [tpl[0] for tpl in match_X],
            doc_counts
        )

    def _create_sparse_matrix(
            self,
            values: Nl,
            concept_names: List[str],
            doc_counts: List[int]
            ) -> csr_matrix:
        return csr_matrix(
            (
                values,
                (
                    [
                        doc_idx
                        for doc_idx
                        in range(len(doc_counts))
                        for _
                        in range(doc_counts[doc_idx])
                        ],
                    [
                        self.concept_map_.get(name)
                        for name
                        in concept_names
                    ]
                )),
            shape=(len(doc_counts), len(self.concept_map_))
        )

    @staticmethod
    def _collect_prediction_results(
            values: Nl,
            concept_names: List[T],
            doc_counts: List[int]
            ) -> List[Tuple[List[T], Nl]]:
        ret = []
        start = 0
        for count in doc_counts:
            end = start+count
            ret.append((concept_names[start:end], values[start:end]))
            start = end
        return ret

    def match_and_extend(
            self,
            inputs: Iterable[str],
            truth_refss: Iterable[Container] = None
            ) -> Tuple[List[Tuple[
                str,
                spmatrix,
                array,
                int,
                List[int], int]],
                List[int]]:
        """Retrieves concepts by their labels from text.
        If ground truth values are present,
        it will also return a list of labels for scoring matches.
        If no ground truth values are present, a list
        with the number of matched concepts for each document is returned."""
        concepts = []
        input_handler = get_input_handler(self.input)
        if truth_refss is not None:
            ret_y = []
            for inp, truth_refs in zip(inputs, map(str, truth_refss)):
                text = input_handler(inp)
                if self.use_txt_vec:
                    txt_vec = self.text_vectorizer_.transform([inp])[0]
                else:
                    txt_vec = 0
                txt_feat = self.text_features_.transform([text])[0]
                matched_concepts: Dict[str, List[int]] = defaultdict(list)
                for match in self.dfa_.search(text):
                    concept = match[0]
                    position = match[2]
                    matched_concepts[concept].append(position)
                for concept, positions in matched_concepts.items():
                    concepts.append((
                        concept,
                        txt_feat,
                        txt_vec,
                        len(text),
                        positions,
                        0))
                    ret_y.append(int(concept in truth_refs))
                self._mark_last_concept_in_doc(concepts)
            return concepts, ret_y
        else:
            doc_counts: List[int] = []
            for inp in inputs:
                text = input_handler(inp)
                if self.use_txt_vec:
                    txt_vec = self.text_vectorizer_.transform([inp])[0]
                else:
                    txt_vec = 0
                txt_feat = self.text_features_.transform([text])[0]
                matched_concepts = defaultdict(list)
                for match in self.dfa_.search(text):
                    concept = match[0]
                    position = match[2]
                    matched_concepts[concept].append(position)
                for concept, positions in matched_concepts.items():
                    concepts.append((
                        concept,
                        txt_feat,
                        txt_vec,
                        len(text),
                        positions,
                        0))
                self._mark_last_concept_in_doc(concepts)
                doc_counts.append(len(matched_concepts))
            return concepts, doc_counts

    def _mark_last_concept_in_doc(self, concepts):
        if concepts:
            last = concepts.pop()
            concepts.append((last[0], last[1], last[2], last[3], last[4], 1))

    def store(self, path):
        with ZipFile(path, 'w') as zfile:
            with zfile.open(_NAME_PREDICTOR_FILE, 'w', force_zip64=True) as fp:
                fp.write(
                    dumps({
                        _KEY_DFA: self.dfa_.to_dict(str),
                        _KEY_CONCEPT_MAP: self.concept_map_,
                        _KEY_CONCEPT_TYPE_URI: _store_uri_ref(
                            self.concept_type_uri),
                        _KEY_THESAURUS_TYPE_URI: _store_uri_ref(
                            self.sub_thesaurus_type_uri),
                        _KEY_THESAURUS_RELATION_TYPE_URI: _store_uri_ref(
                            self.thesaurus_relation_type_uri),
                        _KEY_THESAURUS_RELATION_IS_SPECIALISATION: (
                                self.thesaurus_relation_is_specialisation),
                        _KEY_REMOVE_DEPRECATED: self.remove_deprecated,
                        _KEY_LANGS: list(self.langs),
                        _KEY_INPUT: self.input,
                        _KEY_USE_TXT_VEC: self.use_txt_vec,
                        _KEY_HANDLE_TITLE_CASE: self.handle_title_case,
                        _KEY_EXTRACT_UPPER_CASE_FROM_BRACES: (
                            self.extract_upper_case_from_braces),
                        _KEY_EXTRACT_ANY_CASE_FROM_BRACES: (
                            self.extract_any_case_from_braces),
                        _KEY_EXPAND_AMPERSAND_WITH_SPACES: (
                            self.expand_ampersand_with_spaces),
                        _KEY_EXPAND_ABBREVIATION_WITH_PUNCTUATION: (
                            self.expand_abbreviation_with_punctuation),
                        _KEY_SIMPLE_ENGLISH_PLURAL_RULES: (
                            self.simple_english_plural_rules),
                        },
                        ensure_ascii=False
                    ).encode('utf-8')
                )
            with zfile.open(_NAME_PIPELINE_FILE, 'w', force_zip64=True) as fp:
                # No good way to serialize sk-learn classifier,
                # apart from insecure pickling
                pkl.dump(self.pipeline_, fp)
            if self.use_txt_vec:
                with zfile.open(
                        _NAME_TEXT_VECTORIZER_FILE,
                        'w',
                        force_zip64=True) as fp:
                    pkl.dump(self.text_vectorizer_, fp)
            with zfile.open(
                    _NAME_TEXT_FEATURES_FILE,
                    'w',
                    force_zip64=True) as fp:
                pkl.dump(self.text_features_, fp)
            with zfile.open(_NAME_GRAPH_FILE, 'w', force_zip64=True) as fp:
                fp.write(self.graph.serialize(encoding='utf-8'))

    @staticmethod
    def load(path):
        with ZipFile(path, 'r') as zfile:
            with zfile.open(_NAME_PREDICTOR_FILE, 'r') as fp:
                conf = loads(fp.read().decode('utf-8'))
            use_txt_vec = conf[_KEY_USE_TXT_VEC]
            if use_txt_vec:
                with zfile.open(_NAME_TEXT_VECTORIZER_FILE, 'r') as fp:
                    text_vectorizer = pkl.load(fp)
            with zfile.open(_NAME_GRAPH_FILE, 'r') as fp:
                graph = Graph()
                graph.parse(data=fp.read().decode('utf-8'))
            with zfile.open(_NAME_PIPELINE_FILE, 'r') as fp:
                pipeline = pkl.load(fp)
            with zfile.open(_NAME_TEXT_FEATURES_FILE, 'r') as fp:
                text_features = pkl.load(fp)
        pred = StwfsapyPredictor(
            graph=graph,
            concept_type_uri=_load_uri_ref(
                conf[_KEY_CONCEPT_TYPE_URI]),
            sub_thesaurus_type_uri=_load_uri_ref(
                conf[_KEY_THESAURUS_TYPE_URI]),
            thesaurus_relation_type_uri=_load_uri_ref(
                conf[_KEY_THESAURUS_RELATION_TYPE_URI]),
            thesaurus_relation_is_specialisation=(
                conf[_KEY_THESAURUS_RELATION_IS_SPECIALISATION]),
            remove_deprecated=conf[_KEY_REMOVE_DEPRECATED],
            langs=frozenset(conf[_KEY_LANGS]),
            input=conf[_KEY_INPUT],
            use_txt_vec=use_txt_vec,
            handle_title_case=conf[_KEY_HANDLE_TITLE_CASE],
            extract_upper_case_from_braces=conf[
                _KEY_EXTRACT_UPPER_CASE_FROM_BRACES],
            extract_any_case_from_braces=conf[
                _KEY_EXTRACT_ANY_CASE_FROM_BRACES],
            expand_ampersand_with_spaces=conf[
                _KEY_EXPAND_AMPERSAND_WITH_SPACES],
            expand_abbreviation_with_punctuation=conf[
                _KEY_EXPAND_ABBREVIATION_WITH_PUNCTUATION],
            simple_english_plural_rules=conf[
                _KEY_SIMPLE_ENGLISH_PLURAL_RULES]
        )
        pred.text_features_ = text_features
        if use_txt_vec:
            pred.text_vectorizer_ = text_vectorizer
        else:
            pred.text_vectorizer_ = None
        pred.dfa_ = dfa.Dfa.from_dict(conf[_KEY_DFA], str)
        pred.pipeline_ = pipeline
        pred.concept_map_ = conf[_KEY_CONCEPT_MAP]
        return pred


def _store_uri_ref(ref: URIRef) -> str:
    return ref.toPython()


def _load_uri_ref(uri: str) -> URIRef:
    return URIRef(uri)


def _handle_construction(
        con_state: construction.ConstructionState,
        concept: str,
        label: str):
    """Wrapper for construction that logs a warning
    in case of an exception.
    Uses concept and label as arguments for a more detailed message."""
    try:
        con_state.construct()
    except Exception:
        _logger.warning(
            f'Could not process label "{label}" of concept "{concept}".')
